from typing import Any, Callable, TYPE_CHECKING, Coroutine
from autogen import AssistantAgent, UserProxyAgent
if TYPE_CHECKING:
  from autotx.AutoTx import AutoTx

class AutoTxTool:
  name: str
  description: str

  def register_tool(self, autotx: 'AutoTx', agent: AssistantAgent, user_proxy: UserProxyAgent) -> None:
    tool = self.build_tool(autotx)
    # Register the tool signature with the assistant agent.
    agent.register_for_llm(name=self.name, description=self.description)(tool)
    # Register the tool function with the user proxy agent.
    user_proxy.register_for_execution(name=self.name)(tool)

  def build_tool(self, autotx: 'AutoTx') -> Callable[..., Any] | Callable[..., Coroutine[Any, Any, Any]]:
    raise NotImplementedError


import uuid
from fastapi.testclient import TestClient
import pytest
from web3 import Web3
from autotx import db, server
from autotx.server import app
from fastapi.testclient import TestClient

client = TestClient(app)

@pytest.fixture
def setup():
    db.clear_db()
    db.create_app("test", "1234")
    server.setup_server(verbose=True, logs=None, max_rounds=None, cache=False, is_dev=False, check_valid_safe=False)

def test_connect_auth(setup):
    
    user_id = uuid.uuid4().hex

    response = client.post("/api/v1/connect", json={
        "user_id": user_id,
    })
    assert response.status_code == 401

def test_connect(setup):
    user_id = uuid.uuid4().hex
   
    response = client.post("/api/v1/connect", json={
        "user_id": user_id,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    assert "user_id" in data
    assert "agent_address" in data
    assert "created_at" in data

    assert data["user_id"] == user_id
    assert Web3.is_address(data["agent_address"])

    response2 = client.post("/api/v1/connect", json={
        "user_id": user_id,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response2.status_code == 200
    data2 = response2.json()

    assert "user_id" in data2
    assert "agent_address" in data2
    assert "created_at" in data2

    assert data2["user_id"] == user_id
    assert Web3.is_address(data2["agent_address"])
    assert data2["agent_address"] == data["agent_address"]




from dotenv import load_dotenv

from autotx.utils.configuration import AppConfig
from autotx.utils.ethereum.helpers.swap_from_eoa import swap
from autotx.utils.ethereum.send_native import send_native
from autotx.wallets.safe_smart_wallet import SafeSmartWallet
load_dotenv()

from autotx.agents.DelegateResearchTokensAgent import DelegateResearchTokensAgent
from autotx.agents.SendTokensAgent import SendTokensAgent
from autotx.agents.SwapTokensAgent import SwapTokensAgent

from eth_account import Account

from autotx.utils.constants import OPENAI_API_KEY, OPENAI_MODEL_NAME
from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.helpers.get_dev_account import get_dev_account

import pytest
from autotx.AutoTx import AutoTx, Config
from autotx.chain_fork import stop, start
from autotx.utils.ethereum import (
    transfer_erc20,
)

@pytest.fixture(autouse=True)
def start_and_stop_local_fork():
    start()

    yield

    stop()

@pytest.fixture()
def configuration():
    app_config = AppConfig.load()
    wallet = SafeSmartWallet(app_config.manager, auto_submit_tx=True)
    dev_account = get_dev_account()

    send_native(dev_account, wallet.address, 10, app_config.web3)

    return (dev_account, app_config.agent, app_config.client, app_config.manager, wallet)

@pytest.fixture()
def auto_tx(configuration):
    (_, _, client, _, wallet) = configuration
    network_info = NetworkInfo(client.w3.eth.chain_id)
    get_llm_config = lambda: { "cache_seed": None, "config_list": [{"model": OPENAI_MODEL_NAME, "api_key": OPENAI_API_KEY}]}

    return AutoTx(
        client.w3,
        wallet,
        network_info, 
        [
            SendTokensAgent(),
            SwapTokensAgent(),
            DelegateResearchTokensAgent()
        ], 
        Config(verbose=True, get_llm_config=get_llm_config, logs_dir=None, log_costs=True, max_rounds=100), 
    )

@pytest.fixture()
def usdc(configuration) -> ETHAddress:
    (user, _, client, _, wallet) = configuration
 
    chain_id = client.w3.eth.chain_id
    network_info = NetworkInfo(chain_id)
    
    eth_address = ETHAddress(network_info.tokens["eth"])
    usdc_address = ETHAddress(network_info.tokens["usdc"])

    amount = 100

    swap(client, user, amount, eth_address, usdc_address, network_info.chain_id)
    
    transfer_erc20(client.w3, usdc_address, user, wallet.address, amount)

    return usdc_address

@pytest.fixture()
def test_accounts() -> list[ETHAddress]:
    # Create 10 random test accounts
    accounts = [ETHAddress(Account.create().address) for _ in range(10)]

    return accounts



from eth_typing import ChecksumAddress
from web3 import HTTPProvider, Web3

from autotx.utils.constants import MAINNET_DEFAULT_RPC

class ETHAddress:
    hex: ChecksumAddress
    ens_domain: str | None

    def __init__(self, hex_or_ens: str):
        if hex_or_ens.endswith(".eth"):
            web3 = Web3(HTTPProvider(MAINNET_DEFAULT_RPC))
            address = web3.ens.address(hex_or_ens)  # type: ignore
            if address == None:
                raise ValueError(f"Invalid ENS: {hex_or_ens}")
            self.hex = address
            self.ens_domain = hex_or_ens
        elif Web3.is_address(hex_or_ens):
            self.hex = Web3.to_checksum_address(hex_or_ens)
            self.ens_domain = None
        else:
            raise ValueError(f"Invalid address: {hex_or_ens}")

    def __repr__(self) -> str:
        return f"{self.ens_domain}({self.hex})" if self.ens_domain else self.hex



import os
from typing import Optional

class Cache:
    folder: str = ".cache"

    def __init__(self, folder: Optional[str]):
        if folder:
            self.folder = folder
        os.makedirs(self.folder, exist_ok=True)

    def read(self, file_name: str) -> str | None:
        try:
            with open(os.path.join(self.folder, file_name), "r") as file:
                return file.read().strip()  # Use strip() to remove newline characters
        except FileNotFoundError:
            print(file_name + " not found") 
            raise
        except Exception as e:
            print(f"An error occurred while reading {file_name}: {e}")
            raise

    def write(self, file_name: str, data: str) -> None:
        with open(os.path.join(self.folder, file_name), "w") as f:
            f.write(data)

    def remove(self, file_name: str) -> None:
        try:
            os.remove(os.path.join(self.folder, file_name))
        except FileNotFoundError:
            return
        except Exception as e:
            print(f"An error occurred while deleting {file_name}: {e}")
            raise

cache = Cache(folder=None)



WETH_ABI = [
    {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "allowance", "type": "uint256"},
            {"internalType": "uint256", "name": "needed", "type": "uint256"},
        ],
        "name": "ERC20InsufficientAllowance",
        "type": "error",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "uint256", "name": "balance", "type": "uint256"},
            {"internalType": "uint256", "name": "needed", "type": "uint256"},
        ],
        "name": "ERC20InsufficientBalance",
        "type": "error",
    },
    {
        "inputs": [{"internalType": "address", "name": "approver", "type": "address"}],
        "name": "ERC20InvalidApprover",
        "type": "error",
    },
    {
        "inputs": [{"internalType": "address", "name": "receiver", "type": "address"}],
        "name": "ERC20InvalidReceiver",
        "type": "error",
    },
    {
        "inputs": [{"internalType": "address", "name": "sender", "type": "address"}],
        "name": "ERC20InvalidSender",
        "type": "error",
    },
    {
        "inputs": [{"internalType": "address", "name": "spender", "type": "address"}],
        "name": "ERC20InvalidSpender",
        "type": "error",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"name": "wad", "type": "uint256"}],
        "name": "withdraw",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "deposit",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
]



from autotx.tests.agents.token.research.test_research import get_top_token_addresses_by_market_cap

def test_research_and_buy_one(configuration, auto_tx):
    (_, _, _, manager, _) = configuration
    
    prompt = (
        f"Buy 1 ETH worth of a meme token with the largest market cap in Ethereum mainnet"
    )
    
    auto_tx.run(prompt, non_interactive=True)

    token_address = get_top_token_addresses_by_market_cap("meme-token", "MAINNET", 1, auto_tx)[0]

    token_balance_in_safe = manager.balance_of(token_address)
    assert token_balance_in_safe > 1000

def test_research_and_buy_multiple(configuration, auto_tx):
    (_, _, _, manager, _) = configuration

    old_eth_balance = manager.balance_of()

    prompt = f"""
        Buy 1 ETH worth of a meme token with the largest market cap
        Then buy the governance token with the largest market cap with 0.5 ETH
        This should be on ethereum mainnet
        """
    
    auto_tx.run(prompt, non_interactive=True)
    
    new_eth_balance = manager.balance_of()

    assert old_eth_balance - new_eth_balance == 1.5

    meme_token_address = get_top_token_addresses_by_market_cap("meme-token", "MAINNET", 1, auto_tx)[0]

    governance_token_address = get_top_token_addresses_by_market_cap("governance", "MAINNET", 1, auto_tx)[0]

    meme_token_balance_in_safe = manager.balance_of(meme_token_address)
    assert meme_token_balance_in_safe > 1000

    governance_token_balance_in_safe = manager.balance_of(governance_token_address)
    assert governance_token_balance_in_safe > 90



from web3 import Web3
from web3.types import TxParams, Wei

from autotx.utils.ethereum.eth_address import ETHAddress
from .constants import GAS_PRICE_MULTIPLIER
from .erc20_abi import ERC20_ABI

def build_transfer_erc20(web3: Web3, token_address: ETHAddress, to: ETHAddress, value: float) -> TxParams:
    erc20 = web3.eth.contract(address=token_address.hex, abi=ERC20_ABI)
    decimals = erc20.functions.decimals().call()
    tx: TxParams = erc20.functions.transfer(
        to.hex, int(value * 10**decimals)
    ).build_transaction(
        {
            "gas": None, # type: ignore
            "gasPrice": Wei(int(web3.eth.gas_price * GAS_PRICE_MULTIPLIER))
        }
    )

    return tx



import subprocess
import sys

def run() -> None:
    # Run mypy check
    result = subprocess.run(["mypy", "."], capture_output=True)
    print(result.stdout.decode())
    if result.returncode != 0:
        print(result.stderr.decode())
        print("Type checking failed")
        sys.exit(1)



MASTER_COPY_ADDRESS = "0xd9Db270c1B5E3Bd161E8c8503c55cEABeE709552"
MASTER_COPY_L2_ADDRESSES = ["0xfb1bffC9d739B8D520DaF37dF666da4C687191EA", "0x3E5c63644E683549055b9Be8653de26E0B4CD36E"]
PROXY_FACTORY_ADDRESS = "0xa6B71E26C5e0845f74c812102Ca7114b6a896AB2"
MULTI_SEND_ADDRESS = "0xA238CBeb142c10Ef7Ad8442C6D1f9E89e07e7761"
GAS_PRICE_MULTIPLIER = 1.1
DEVNET_RPC_URL = "http://localhost:8545"
NATIVE_TOKEN_ADDRESS = "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"



from typing import Any, Optional, cast
import re
import logging

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from gnosis.eth import EthereumClient
from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.multicall import Multicall
from gnosis.safe import Safe, SafeOperation, SafeTx
from gnosis.safe.multi_send import MultiSend, MultiSendOperation, MultiSendTx
from web3.types import TxParams, TxReceipt
from gnosis.safe.api import TransactionServiceApi
from eth_account.signers.local import LocalAccount

from autotx import models
from autotx.transactions import Transaction
from autotx.utils.ethereum.get_native_balance import get_native_balance
from autotx.utils.ethereum.cached_safe_address import get_cached_safe_address, save_cached_safe_address
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.is_valid_safe import is_valid_safe
from autotx.utils.ethereum.networks import ChainId, NetworkInfo
from .deploy_safe_with_create2 import deploy_safe_with_create2
from .deploy_multicall import deploy_multicall
from .get_erc20_balance import get_erc20_balance
from .constants import MULTI_SEND_ADDRESS, GAS_PRICE_MULTIPLIER


# Disable safe warning logs
logging.getLogger('gnosis.safe.safe').setLevel(logging.CRITICAL + 1)

class ExecutionRevertedError(Exception):
    pass

class SafeManager:
    multisend: MultiSend | None = None
    safe_nonce: int | None = None
    gas_multiplier: float | None = GAS_PRICE_MULTIPLIER
    dev_account: LocalAccount | None = None
    network: NetworkInfo
    transaction_service_url: str | None = None
    address: ETHAddress
    use_tx_service: bool

    def __init__(
        self, 
        client: EthereumClient, 
        agent: LocalAccount, 
        safe: Safe
    ):
        self.client = client
        self.web3 = self.client.w3
        self.agent = agent
        self.safe = safe
        self.use_tx_service = False
        self.safe_nonce = None
        self.network = NetworkInfo(client.w3.eth.chain_id)
        self.address = ETHAddress(safe.address)


    @classmethod
    def deploy_safe(
        cls, 
        client: EthereumClient, 
        dev_account: LocalAccount, 
        agent: LocalAccount, 
        owners: list[str], 
        threshold: int
    ) -> 'SafeManager':
        safe_address = get_cached_safe_address()
        if not safe_address:
            safe_address = deploy_safe_with_create2(client, dev_account, owners, threshold)
            save_cached_safe_address(safe_address)

        manager = cls(client, agent, Safe(Web3.to_checksum_address(safe_address), client))
        manager.dev_account = dev_account

        manager.multisend = MultiSend(client, address=Web3.to_checksum_address(MULTI_SEND_ADDRESS))

        return manager
    
    @classmethod
    def connect(
        cls, 
        client: EthereumClient, 
        safe_address: ETHAddress,
        agent: LocalAccount, 
    ) -> 'SafeManager':
        safe = Safe(Web3.to_checksum_address(safe_address.hex), client)

        manager = cls(client, agent, safe)

        manager.multisend = MultiSend(client, address=Web3.to_checksum_address(MULTI_SEND_ADDRESS))

        return manager
    
    def connect_tx_service(self, transaction_service_url: str) -> None:
        self.use_tx_service = True
        self.transaction_service_url = transaction_service_url
    
    def disconnect_tx_service(self) -> None:
        self.use_tx_service = False
        self.transaction_service_url = None

    def connect_multisend(self, address: ChecksumAddress) -> None:
        self.multisend = MultiSend(self.client, address=address)

    def connect_multicall(self, address: ETHAddress) -> None:
        self.client.multicall = Multicall(self.client, address.hex)

    def deploy_multicall(self) -> None:
        if not self.dev_account:
            raise ValueError("Dev account not set. This function should not be called in production.")
        multicall_addr = deploy_multicall(self.client, self.dev_account)
        self.connect_multicall(multicall_addr)

    def build_multisend_tx(self, txs: list[TxParams | dict[str, Any]], safe_nonce: Optional[int] = None) -> SafeTx:
        if not self.multisend:
            raise Exception("No multisend contract address has been set to SafeManager")

        multisend_txs = [
            MultiSendTx(MultiSendOperation.CALL, str(tx["to"]), tx["value"], tx["data"])
            for tx in txs
        ]
        safe_multisend_data = self.multisend.build_tx_data(multisend_txs)


        safe_tx = self.safe.build_multisig_tx(
            to=str(self.multisend.address),
            value=sum(tx["value"] for tx in txs),
            data=safe_multisend_data,
            operation=SafeOperation.DELEGATE_CALL.value,
            safe_nonce=self.track_nonce(safe_nonce)
        )

        return safe_tx

    def build_tx(self, tx: TxParams | dict[str, Any], safe_nonce: Optional[int] = None, skip_estimate_gas: bool = False) -> SafeTx:
        safe_tx = SafeTx(
            self.client,
            self.address.hex,
            str(tx["to"]),
            tx["value"],
            cast(bytes, tx["data"]),
            0,
            0,
            0,
            self.gas_price(),
            None,
            self.address.hex,
            safe_nonce=self.track_nonce(safe_nonce),
        )

        if not skip_estimate_gas:
            safe_tx.safe_tx_gas = self.safe.estimate_tx_gas(safe_tx.to, safe_tx.value, safe_tx.data, safe_tx.operation)
            safe_tx.base_gas = self.safe.estimate_tx_base_gas(safe_tx.to, safe_tx.value, safe_tx.data, safe_tx.operation, NULL_ADDRESS, safe_tx.safe_tx_gas)

        return safe_tx
    
    def execute_tx(self, tx: TxParams | dict[str, Any], safe_nonce: Optional[int] = None) -> HexBytes:
        if not self.dev_account:
            raise ValueError("Dev account not set. This function should not be called in production.")

        try:
            safe_tx = self.build_tx(tx, safe_nonce)

            safe_tx.sign(self.agent.key.hex())

            safe_tx.call(tx_sender_address=self.dev_account.address)

            tx_hash, _ = safe_tx.execute(
                tx_sender_private_key=self.dev_account.key.hex()
            )

            print(f"Executed safe tx hash: {tx_hash.hex()}")

            return tx_hash
        except Exception as e:
            extracted_message = re.search(r"revert: ([^,]+)", str(e))
            if extracted_message:
                raise ExecutionRevertedError(extracted_message.group(0))
            
            raise Exception("Unknown error executing transaction", e)


    def execute_multisend_tx(self, txs: list[TxParams | dict[str, Any]], safe_nonce: Optional[int] = None) -> HexBytes:
        if not self.dev_account:
            raise ValueError("Dev account not set. This function should not be called in production.")

        safe_tx = self.build_multisend_tx(txs, safe_nonce)

        safe_tx.sign(self.agent.key.hex())

        safe_tx.call(tx_sender_address=self.dev_account.address)

        tx_hash, _ = safe_tx.execute(
            tx_sender_private_key=self.dev_account.key.hex()
        )

        return tx_hash
    
    def post_transaction(self, tx: TxParams | dict[str, Any], safe_nonce: Optional[int] = None) -> None:
        ts_api = TransactionServiceApi(
            self.network.chain_id, ethereum_client=self.client, base_url=self.transaction_service_url
        )

        safe_tx = self.build_tx(tx, safe_nonce, skip_estimate_gas=True)
        safe_tx.sign(self.agent.key.hex())

        ts_api.post_transaction(safe_tx)

    def post_multisend_transaction(self, txs: list[TxParams | dict[str, Any]], safe_nonce: Optional[int] = None) -> None:
        ts_api = TransactionServiceApi(
            self.network.chain_id, ethereum_client=self.client, base_url=self.transaction_service_url
        )

        tx = self.build_multisend_tx(txs, safe_nonce)
        tx.sign(self.agent.key.hex())

        ts_api.post_transaction(tx)

    def send_tx(self, tx: TxParams | dict[str, Any], safe_nonce: Optional[int] = None) -> str | None:
        if self.use_tx_service:
            self.post_transaction(cast(TxParams, tx), safe_nonce)
            return None
        else:
            hash = self.execute_tx(cast(TxParams, tx), safe_nonce)
            return hash.hex()
        
    def send_multisend_tx(self, txs: list[TxParams | dict[str, Any]], safe_nonce: Optional[int] = None) -> str | None:
        if self.use_tx_service:
            if len(txs) == 1:
                self.post_transaction(txs[0], safe_nonce)
            elif len(txs) > 1:
                self.post_multisend_transaction(txs, safe_nonce)
            return None
        else:
            if len(txs) == 1:
                hash = self.execute_tx(txs[0], safe_nonce)
            elif len(txs) > 1:
                hash = self.execute_multisend_tx(txs, safe_nonce)
            return hash.hex()
        
    def send_tx_batch(self, txs: list[Transaction], require_approval: bool, safe_nonce: Optional[int] = None) -> bool | str: # True if sent, False if declined, str if feedback
        print("=" * 50)

        if not txs:
            print("No transactions to send.")
            return True

        start_nonce = self.track_nonce(safe_nonce)

        transactions_info = "\n".join(
            [
                f"{i + 1}. {tx.summary} (nonce: {start_nonce + i})"
                for i, tx in enumerate(txs)
            ]
        )

        print(f"Prepared transactions:\n{transactions_info}")

        if self.use_tx_service:
            if require_approval:
                response = input("Do you want the above transactions to be sent to your smart account?\nRespond (y/n) or write feedback: ")

                if response.lower() == "n" or response.lower() == "no":
                    print("Transactions not sent to your smart account (declined).")
                  
                    self.reset_nonce(start_nonce)
                  
                    return False
                elif response.lower() != "y" and response.lower() != "yes":
                    
                    self.reset_nonce(start_nonce)
                    
                    return response
            else:
                print("Non-interactive mode enabled. Transactions will be sent to your smart account without approval.")

            print("Sending transactions to your smart account...")

            for i, tx in enumerate([prepared_tx.params for prepared_tx in txs]):
                self.send_tx(tx, start_nonce + i)

            print("Transactions sent to your smart account for signing.")
            
            return True
        else:
            if require_approval:
                response = input("Do you want to execute the above transactions?\nRespond (y/n) or write feedback: ")

                if response.lower() == "n" or response.lower() == "no":
                    print("Transactions not executed (declined).")
                    
                    self.reset_nonce(start_nonce)
                    
                    return False
                elif response.lower() != "y" and response.lower() != "yes":
                    
                    self.reset_nonce(start_nonce)
                    
                    return response
            else:
                print("Non-interactive mode enabled. Transactions will be executed without approval.")

            print("Executing transactions...")

            for i, prepared_tx in enumerate([prepared_tx for prepared_tx in txs]):
                try:
                    self.send_tx(prepared_tx.params, start_nonce + i)
                except ExecutionRevertedError as e:
                    raise Exception(f"{prepared_tx.summary} failed with error: {e}")
        
            print("Transactions executed.")

            return True

    def send_multisend_tx_batch(self, txs: list[Transaction], require_approval: bool, safe_nonce: Optional[int] = None) -> bool | str: # True if sent, False if declined, str if feedback
        print("=" * 50)

        if not txs:
            print("No transactions to send.")
            return True

        transactions_info = "\n".join(
            [
                f"{i + 1}. {tx.summary}"
                for i, tx in enumerate(txs)
            ]
        )

        print(f"Prepared transactions:\n{transactions_info}")

        if self.use_tx_service:
            if require_approval:
                response = input("Do you want the above transactions to be sent to your smart account?\nRespond (y/n) or write feedback: ")

                if response.lower() == "n" or response.lower() == "no":
                    print("Transactions not sent to your smart account (declined).")
                  
                    return False
                elif response.lower() != "y" and response.lower() != "yes":
                    
                    return response
            else:
                print("Non-interactive mode enabled. Transactions will be sent to your smart account without approval.")

            print("Sending batched transactions to your smart account...")

            self.send_multisend_tx([prepared_tx.params for prepared_tx in txs], safe_nonce)

            if len(txs) == 1:
                print("Transaction sent to your smart account for signing.")
            else:
                print("Transactions sent as a single multi-send transaction to your smart account for signing.")
            
            return True
        else:
            if require_approval:
                response = input("Do you want to execute the above transactions?\nRespond (y/n) or write feedback: ")

                if response.lower() == "n" or response.lower() == "no":
                    print("Transactions not executed (declined).")
                    
                    return False
                elif response.lower() != "y" and response.lower() != "yes":
                    
                    return response
            else:
                print("Non-interactive mode enabled. Transactions will be executed without approval.")

            print("Executing transactions...")

            try:
                self.send_multisend_tx([prepared_tx.params for prepared_tx in txs], safe_nonce)
            except ExecutionRevertedError as e:
                raise Exception(f"Executing transactions failed with error: {e}")
        
            if len(txs) == 1:
                print("Transaction executed.")
            else:
                print("Transactions executed as a single multi-send transaction.")

            return True

    def send_empty_tx(self, safe_nonce: Optional[int] = None) -> str | None:
        tx: TxParams = {
            "to": self.address.hex,
            "value": self.web3.to_wei(0, "ether"),
            "data": b"",
            "from": self.address.hex,
        }

        return self.send_tx(tx, safe_nonce)

    def wait(self, tx_hash: HexBytes) -> TxReceipt:
        return self.web3.eth.wait_for_transaction_receipt(tx_hash)

    def balance_of(self, token_address: ETHAddress | None = None) -> float:
        if token_address is None:
            return get_native_balance(self.web3, self.address)
        else:
            return get_erc20_balance(self.web3, token_address, self.address) 
        
    def nonce(self) -> int:
        return self.safe.retrieve_nonce()

    def gas_price(self) -> int:
        return self.web3.eth.gas_price if self.gas_multiplier is None else int(self.web3.eth.gas_price * self.gas_multiplier)

    def track_nonce(self, safe_nonce: Optional[int] = None) -> int:
        if safe_nonce is None:
            if self.safe_nonce is None:
                self.safe_nonce = self.nonce()
            else:
                self.safe_nonce += 1
            return self.safe_nonce
        else:
            return safe_nonce
    
    def reset_nonce(self, starting_safe_nonce: Optional[int] = None) -> None:
        if starting_safe_nonce is None:
            self.safe_nonce = None
        else:
            self.safe_nonce = starting_safe_nonce - 1 # -1 because it will be incremented in track_nonce

    @staticmethod
    def is_valid_safe(client: EthereumClient, address: ETHAddress) -> bool:
        return is_valid_safe(client, address)



from autotx.utils.ethereum.cache import cache

SAFE_ADDRESS_FILE_NAME = "safe.txt"

def get_cached_safe_address() -> str | None:
    try:
        return cache.read(SAFE_ADDRESS_FILE_NAME)
    except Exception as e:
        return None

def save_cached_safe_address(safe_address: str) -> None:
    cache.write(SAFE_ADDRESS_FILE_NAME, safe_address)

def delete_cached_safe_address() -> None:
    cache.remove(SAFE_ADDRESS_FILE_NAME)



from eth_account.signers.local import LocalAccount
from hexbytes import HexBytes
from web3 import Web3
from web3.types import Wei
from web3.middleware.signing import construct_sign_and_send_raw_middleware

from autotx.utils.ethereum.eth_address import ETHAddress

from .constants import GAS_PRICE_MULTIPLIER

from .erc20_abi import ERC20_ABI


def transfer_erc20(
    web3: Web3,
    token_address: ETHAddress,
    from_account: LocalAccount,
    to: ETHAddress,
    value: float,
) -> HexBytes:
    account_middleware = construct_sign_and_send_raw_middleware(from_account)
    web3.middleware_onion.add(account_middleware)

    erc20 = web3.eth.contract(address=token_address.hex, abi=ERC20_ABI)
    decimals = erc20.functions.decimals().call()

    tx_hash = erc20.functions.transfer(to.hex, int(value * 10**decimals)).transact(
        {
            "from": from_account.address,
            "gasPrice": Wei(int(web3.eth.gas_price * GAS_PRICE_MULTIPLIER)),
        }
    )

    web3.middleware_onion.remove(account_middleware) # type: ignore

    return tx_hash



from eth_typing import HexStr
from gnosis.eth import EthereumClient
from web3 import Web3
from gnosis.safe import Safe

from autotx.utils.ethereum.eth_address import ETHAddress
from .constants import MASTER_COPY_ADDRESS, MASTER_COPY_L2_ADDRESSES

def is_valid_safe(client: EthereumClient, safe_address: ETHAddress) -> bool:
    w3 = client.w3

    if w3.eth.get_code(Web3.to_checksum_address(safe_address.hex)) != w3.to_bytes(hexstr=HexStr("0x")):
        safe = Safe(Web3.to_checksum_address(safe_address.hex), client)
        master_copy_address = safe.retrieve_master_copy_address()
        return master_copy_address in [MASTER_COPY_ADDRESS] or master_copy_address in MASTER_COPY_L2_ADDRESSES
    else:
        return False


from datetime import datetime
import os
from typing import Any, Callable, Dict, Optional
from eth_account.signers.local import LocalAccount
from gnosis.eth import EthereumClient

from autotx.agents.DelegateResearchTokensAgent import DelegateResearchTokensAgent
from autotx.agents.SendTokensAgent import SendTokensAgent
from autotx.agents.SwapTokensAgent import SwapTokensAgent
from autotx.autotx_agent import AutoTxAgent
from autotx.utils.constants import COINGECKO_API_KEY, OPENAI_API_KEY, OPENAI_BASE_URL, OPENAI_MODEL_NAME
from autotx.utils.ethereum import SafeManager
from autotx.utils.ethereum.cached_safe_address import get_cached_safe_address
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.helpers.fill_dev_account_with_tokens import fill_dev_account_with_tokens
from autotx.utils.ethereum.helpers.get_dev_account import get_dev_account
from autotx.utils.ethereum.helpers.show_address_balances import show_address_balances
from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.is_dev_env import is_dev_env
from autotx.utils.ethereum.agent_account import get_or_create_agent_account

def print_agent_address() -> None:
    acc = get_or_create_agent_account()
    print(f"Agent address: {acc.address}")

def setup_safe(smart_account_addr: ETHAddress | None, agent: LocalAccount, client: EthereumClient, fill_dev_account: bool, check_valid_safe: bool) -> SafeManager:
    web3 = client.w3

    chain_id = web3.eth.chain_id

    network_info = NetworkInfo(chain_id)

    if is_dev_env():
        print(f"Connected to fork of {network_info.chain_id.name} network.")
    else:
        print(f"Connected to {network_info.chain_id.name} network.")

    print_agent_address()

    manager: SafeManager

    if smart_account_addr:
        if check_valid_safe and not SafeManager.is_valid_safe(client, smart_account_addr):
            raise Exception(f"Invalid safe address: {smart_account_addr}")

        print(f"Smart account connected: {smart_account_addr}")
        manager = SafeManager.connect(client, smart_account_addr, agent)

        manager.connect_tx_service(network_info.transaction_service_url)
    else:
        print("No smart account connected, deploying a new one...")
        dev_account = get_dev_account()

        is_safe_deployed = get_cached_safe_address()
        manager = SafeManager.deploy_safe(client, dev_account, agent, [dev_account.address, agent.address], 1)
        print(f"Smart account deployed: {manager.address}")
        
        if not is_safe_deployed and fill_dev_account:
            fill_dev_account_with_tokens(client, dev_account, manager.address, network_info)
            print(f"Funds sent to smart account for testing purposes")

        print("=" * 50)
        print("Starting smart account balances:")
        show_address_balances(web3, network_info.chain_id, manager.address)
        print("=" * 50)

    return manager

def setup_agents(logs: str | None, cache: bool | None) -> tuple[Callable[[], Optional[Dict[str, Any]]], list[AutoTxAgent], str | None]:
    now = datetime.now()
    now_str = now.strftime('%Y-%m-%d-%H-%M-%S-') + str(now.microsecond)
    logs_dir = os.path.join(logs, now_str) if logs is not None else None

    get_llm_config = lambda: { 
        "cache_seed": 1 if cache else None, 
        "config_list": [
            {
                "model": OPENAI_MODEL_NAME, 
                "api_key": OPENAI_API_KEY,
                "base_url": OPENAI_BASE_URL
            }
        ]
    }
    agents = [
        SendTokensAgent(),
        SwapTokensAgent()
    ]

    if COINGECKO_API_KEY:
        agents.append(DelegateResearchTokensAgent())

    return (get_llm_config, agents, logs_dir)


from abc import abstractmethod
from decimal import Decimal
from enum import Enum
from pydantic import BaseModel
from typing import Any, Union, cast
from web3.types import TxParams

from web3 import Web3

from autotx.token import Token
from autotx.transactions import SendTransaction, Transaction
from autotx.utils.ethereum.build_transfer_erc20 import build_transfer_erc20
from autotx.utils.ethereum.build_transfer_native import build_transfer_native
from autotx.utils.ethereum.constants import NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.lifi.swap import build_swap_transaction
from autotx.utils.ethereum.networks import NetworkInfo

class IntentType(str, Enum):
    SEND = "send"
    BUY = "buy"
    SELL = "sell"

class IntentBase(BaseModel):
    type: IntentType
    summary: str

    @abstractmethod
    def build_transactions(self, web3: Web3, network: NetworkInfo, smart_wallet_address: ETHAddress) -> list[Transaction]:
        pass

class SendIntent(IntentBase):
    receiver: str
    token: Token
    amount: float

    @classmethod
    def create(cls, token: Token, amount: float, receiver: ETHAddress) -> 'SendIntent':
        return cls(
            type=IntentType.SEND,
            token=token,
            amount=amount,
            receiver=receiver.hex,
            summary=f"Transfer {amount} {token.symbol} to {receiver}",
        )
    
    def build_transactions(self, web3: Web3, network: NetworkInfo, smart_wallet_address: ETHAddress) -> list[Transaction]:
        tx: TxParams

        if self.token.address == NATIVE_TOKEN_ADDRESS:
            tx = build_transfer_native(web3, ETHAddress(self.token.address), ETHAddress(self.receiver), self.amount)
        else:
            tx = build_transfer_erc20(web3, ETHAddress(self.token.address), ETHAddress(self.receiver), self.amount)
            
        transactions: list[Transaction] = [
            SendTransaction.create(
                token=self.token,
                amount=self.amount,
                receiver=self.receiver,
                params=cast(dict[str, Any], tx),
            )
        ]

        return transactions

class BuyIntent(IntentBase):
    from_token: Token
    to_token: Token
    amount: float

    @classmethod
    def create(cls, from_token: Token, to_token: Token, amount: float) -> 'BuyIntent':
        return cls(
            type=IntentType.BUY,
            from_token=from_token,
            to_token=to_token,
            amount=amount,
            summary=f"Buy {amount} {to_token.symbol} with {from_token.symbol}",
        )
    
    def build_transactions(self, web3: Web3, network: NetworkInfo, smart_wallet_address: ETHAddress) -> list[Transaction]:
        transactions = build_swap_transaction(
            web3,
            Decimal(self.amount),
            ETHAddress(self.from_token.address),
            ETHAddress(self.to_token.address),
            smart_wallet_address,
            False,
            network.chain_id
        )

        return transactions

class SellIntent(IntentBase):
    from_token: Token
    to_token: Token
    amount: float

    @classmethod
    def create(cls, from_token: Token, to_token: Token, amount: float) -> 'SellIntent':
        return cls(
            type=IntentType.SELL,
            from_token=from_token,
            to_token=to_token,
            amount=amount,
            summary=f"Sell {amount} {from_token.symbol} for {to_token.symbol}",
        )    
    
    def build_transactions(self, web3: Web3, network: NetworkInfo, smart_wallet_address: ETHAddress) -> list[Transaction]:
        transactions = build_swap_transaction(
            web3,
            Decimal(self.amount),
            ETHAddress(self.from_token.address),
            ETHAddress(self.to_token.address),
            smart_wallet_address,
            True,
            network.chain_id
        )

        return transactions

Intent = Union[SendIntent, BuyIntent, SellIntent]



import json
from textwrap import dedent
from typing import Union

import requests

KLEROS_TOKENS_LIST = "https://t2crtokens.eth.link/"
COINGECKO_TOKENS_LISTS = [
    "https://tokens.coingecko.com/uniswap/all.json",
    "https://tokens.coingecko.com/optimistic-ethereum/all.json",
    "https://tokens.coingecko.com/polygon-pos/all.json",
    "https://tokens.coingecko.com/base/all.json",
    "https://tokens.coingecko.com/arbitrum-one/all.json",
    "https://tokens.coingecko.com/xdai/all.json",
    "https://tokens.coingecko.com/zksync/all.json",
]


TOKENS_LIST = [KLEROS_TOKENS_LIST, *COINGECKO_TOKENS_LISTS]


def fetch_tokens_list() -> None:
    loaded_tokens: list[dict[str, Union[str, int]]] = []

    for token_list_url in TOKENS_LIST:
        try:
            response = requests.get(token_list_url)
            tokens = json.loads(response.text)["tokens"]
            loaded_tokens.extend(tokens)
        except:
            print("Error while trying to fetch list:", token_list_url)

    loaded_tokens_as_string = json.dumps(loaded_tokens, indent=2)
    content = dedent(
        f"""# THIS IS AN AUTOGENERATED FILE. DO NOT MODIFY MANUALLY, PLEASE CHECK autotx/load_tokens.py TO DO ANY MODIFICATIONS

token_list = {loaded_tokens_as_string}
"""
    )
    with open("autotx/utils/ethereum/helpers/token_list.py", "w") as f:
        f.write(content)


def run() -> None:
    fetch_tokens_list()



from web3 import Web3

from autotx.utils.ethereum.eth_address import ETHAddress
from .erc20_abi import ERC20_ABI

def get_erc20_info(web3: Web3, token_address: ETHAddress) -> tuple[str, str, int]:
    erc20 = web3.eth.contract(address=token_address.hex, abi=ERC20_ABI)

    name = erc20.functions.name().call()
    symbol = erc20.functions.symbol().call()
    decimals = erc20.functions.decimals().call()

    return name, symbol, decimals


from web3 import Web3
from web3.types import TxParams

from autotx.utils.ethereum.eth_address import ETHAddress

def build_transfer_native(web3: Web3, from_address: ETHAddress, to: ETHAddress, value: float) -> TxParams:
    return {
        "to": to.hex,
        "value": web3.to_wei(value, "ether"),
        "data": b"",
        "from": from_address.hex,
    }



from autotx.agents.ResearchTokensAgent import (
    filter_token_list_by_network,
    get_coingecko,
)

from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.networks import ChainId

def get_top_token_addresses_by_market_cap(category: str, network: str, count: int, auto_tx) -> list[ETHAddress]:
    tokens = get_coingecko().coins.get_markets(vs_currency="usd", category=category, per_page=250)
    tokens_in_network = filter_token_list_by_network(
        tokens, network
    )

    return [ETHAddress(auto_tx.network.tokens[token["symbol"].lower()]) for token in tokens_in_network[:count]]

def test_price_change_information(auto_tx):
    token_information = get_coingecko().coins.get_id(
        id="starknet",
        localization=False,
        tickers=False,
        community_data=False,
        sparkline=False,
    )
    prompt = "What's the 24 hours price change of Starknet (STRK)?"

    result = auto_tx.run(prompt, non_interactive=True)

    price_change = token_information["market_data"]["price_change_percentage_24h"]
    price_change_rounded = round(price_change, 2)

    assert (
        str(price_change) in "\n".join(result.info_messages).lower() or str(price_change_rounded) in "\n".join(result.info_messages).lower()
    )

def test_get_top_5_tokens_from_base(auto_tx):
    tokens = get_coingecko().coins.get_markets(
        vs_currency="usd", category="base-ecosystem"
    )
    prompt = "What are the top 5 tokens on Base chain?"

    result = auto_tx.run(prompt, non_interactive=True)

    for token in tokens[:5]:
        symbol: str = token["symbol"]
        assert symbol.lower() in "\n".join(result.info_messages).lower()

def test_get_top_5_most_traded_tokens_from_l1(auto_tx):
    tokens = get_coingecko().coins.get_markets(
        vs_currency="usd", category="layer-1", order="volume_desc"
    )
    prompt = "What are the top 5 most traded Layer 1 tokens on all networks?"
    
    result = auto_tx.run(prompt, non_interactive=True)

    for token in tokens[:5]:
        symbol: str = token["symbol"]
        assert symbol.lower() in "\n".join(result.info_messages).lower()

def test_get_top_5_memecoins(auto_tx):
    tokens = get_coingecko().coins.get_markets(vs_currency="usd", category="meme-token")
    tokens_in_network = filter_token_list_by_network(
        tokens, "MAINNET"
    )

    prompt = "What are the top 5 meme coins on my network?"

    result = auto_tx.run(prompt, non_interactive=True)

    for token in tokens_in_network[:5]:
        symbol: str = token["symbol"]
        assert symbol.lower() in "\n".join(result.info_messages).lower()

def test_get_top_5_memecoins_in_optimism(auto_tx):
    tokens = get_coingecko().coins.get_markets(vs_currency="usd", category="meme-token")
    prompt = "What are the top 5 meme coins on Optimism?"

    result = auto_tx.run(prompt, non_interactive=True)

    tokens = filter_token_list_by_network(tokens, ChainId.OPTIMISM.name)
    for token in tokens[:5]:
        symbol: str = token["symbol"]
        assert symbol.lower() in "\n".join(result.info_messages).lower()



from autotx.utils.ethereum import get_erc20_balance, get_native_balance
from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.ethereum.eth_address import ETHAddress

DIFFERENCE_PERCENTAGE = 1.01

def test_swap_and_send_simple(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    receiver = test_accounts[0]

    prompt = f"Swap ETH to 0.05 WBTC, and send 0.01 WBTC to {receiver}"

    auto_tx.run(prompt, non_interactive=True)

    new_wbtc_safe_address = manager.balance_of(wbtc_address)
    new_receiver_wbtc_balance = get_erc20_balance(client.w3, wbtc_address, receiver)
    excepted_safe_wbtc_balance = 0.04
    assert excepted_safe_wbtc_balance <= new_wbtc_safe_address <= new_wbtc_safe_address * DIFFERENCE_PERCENTAGE
    assert new_receiver_wbtc_balance == 0.01

def test_swap_and_send_complex(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    receiver = test_accounts[0]

    prompt = f"Swap ETH to 0.05 WBTC, then, swap WBTC to 1000 USDC and send 50 USDC to {receiver}"

    wbtc_safe_address = manager.balance_of(wbtc_address)
    auto_tx.run(prompt, non_interactive=True)

    new_wbtc_safe_address = manager.balance_of(wbtc_address)
    new_usdc_safe_address = manager.balance_of(usdc_address)
    new_receiver_usdc_balance = get_erc20_balance(client.w3, usdc_address, receiver)

    expected_usdc_safe_balance = 950
    assert new_wbtc_safe_address > wbtc_safe_address
    assert expected_usdc_safe_balance <= new_usdc_safe_address <= expected_usdc_safe_balance * DIFFERENCE_PERCENTAGE
    assert new_receiver_usdc_balance == 50

def test_send_and_swap_simple(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    receiver = test_accounts[0]

    prompt = f"Send 0.1 ETH to {receiver}, and then swap ETH to 0.05 WBTC"

    receiver_native_balance = get_native_balance(client.w3, receiver)
    receiver_wbtc_balance = get_erc20_balance(client.w3, wbtc_address, receiver)

    auto_tx.run(prompt, non_interactive=True)

    safe_wbtc_balance = manager.balance_of(wbtc_address)
    new_receiver_native_balance = get_native_balance(client.w3, receiver)
    new_receiver_wbtc_balance = get_erc20_balance(client.w3, wbtc_address, receiver)

    expected_wbtc_safe_balance = 0.05
    assert expected_wbtc_safe_balance <= safe_wbtc_balance <= expected_wbtc_safe_balance * DIFFERENCE_PERCENTAGE
    assert receiver_wbtc_balance == 0
    assert new_receiver_wbtc_balance == receiver_wbtc_balance
    assert new_receiver_native_balance == receiver_native_balance + 0.1

def test_send_and_swap_complex(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    receiver_1 = test_accounts[0]
    receiver_2 = test_accounts[1]

    prompt = f"Send 0.1 ETH to {receiver_1}, then swap ETH to 0.05 WBTC, then, swap WBTC to 1000 USDC and send 50 USDC to {receiver_2}"

    wbtc_safe_balance = manager.balance_of(wbtc_address)
    receiver_1_native_balance = get_native_balance(client.w3, receiver_1)
    receiver_2_usdc_balance = get_erc20_balance(client.w3, usdc_address, receiver_2)

    auto_tx.run(prompt, non_interactive=True)

    new_wbtc_safe_balance = manager.balance_of(wbtc_address)
    new_usdc_safe_balance = manager.balance_of(usdc_address)
    new_receiver_1_native_balance = get_native_balance(client.w3, receiver_1)
    new_receiver_1_usdc_balance = get_erc20_balance(client.w3, usdc_address, receiver_1)
    new_receiver_1_wbtc_balance = get_erc20_balance(client.w3, wbtc_address, receiver_1)
    new_receiver_2_wbtc_balance = get_erc20_balance(client.w3, wbtc_address, receiver_2)
    new_receiver_2_usdc_balance = get_erc20_balance(client.w3, usdc_address, receiver_2)
    expected_usdc_safe_balance = 950
    assert expected_usdc_safe_balance <= new_usdc_safe_balance <= expected_usdc_safe_balance * DIFFERENCE_PERCENTAGE
    assert new_wbtc_safe_balance > wbtc_safe_balance
    assert new_receiver_1_native_balance == receiver_1_native_balance + 0.1
    assert new_receiver_1_usdc_balance == 0
    assert new_receiver_1_wbtc_balance == 0
    assert new_receiver_2_usdc_balance == receiver_2_usdc_balance + 50
    assert new_receiver_2_wbtc_balance == 0


import uuid
from fastapi.testclient import TestClient
import pytest
from autotx import db, server
from autotx.server import app
from fastapi.testclient import TestClient

from autotx.utils.ethereum.cached_safe_address import get_cached_safe_address

client = TestClient(app)

@pytest.fixture(scope="session", autouse=True)
def setup():
    db.clear_db()
    db.create_app("test", "1234")

def test_create_task_auth():
    smart_wallet_address = get_cached_safe_address()

    response = client.post("/api/v1/tasks", json={
        "prompt": "Send 1 ETH to vitalik.eth",
        "address": smart_wallet_address,
        "chain_id": 1,
        "user_id": "123"
    })
    assert response.status_code == 401

def test_get_tasks_auth():
    response = client.get("/api/v1/tasks")
    assert response.status_code == 401

def test_get_task_auth():
    response = client.get("/api/v1/tasks/123")
    assert response.status_code == 401

def test_create_task():
    server.setup_server(verbose=True, logs=None, max_rounds=None, cache=False, is_dev=True, check_valid_safe=False)
    
    smart_wallet_address = get_cached_safe_address()
    user_id = uuid.uuid4().hex
   
    response = client.post("/api/v1/connect", json={
        "user_id": user_id,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200

    response = client.post("/api/v1/tasks", json={
        "prompt": "Send 1 ETH to vitalik.eth",
        "address": smart_wallet_address,
        "chain_id": 1,
        "user_id": user_id
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    assert "id" in data
    assert "created_at" in data
    assert "updated_at" in data
    assert data["messages"] == []
    assert data["intents"] == []
    assert data["running"] is True

    response = client.get(f"/api/v1/tasks/{data['id']}", headers={
        "Authorization": f"Bearer 1234"
    })
    data = response.json()

    assert data["running"] is False
    assert len(data["intents"]) > 0

def test_get_tasks():
    response = client.get("/api/v1/tasks", headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()
    assert len(data) > 0
    assert "id" in data[0]

def test_get_task():
    response = client.get("/api/v1/tasks", headers={
        "Authorization": f"Bearer 1234"
    })
    task_id = response.json()[0]["id"]
    
    response = client.get(f"/api/v1/tasks/{task_id}", headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()
    assert data["id"] == task_id
    assert len(data["messages"]) > 0
    assert data["running"] is False

def test_get_task_intents():
    response = client.get("/api/v1/tasks", headers={
        "Authorization": f"Bearer 1234"
    })
    task_id = response.json()[0]["id"]

    response = client.get(f"/api/v1/tasks/{task_id}/intents", headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()
    assert len(data) == 1



from io import StringIO, TextIOWrapper
from typing import TextIO

class TeeStream(TextIOWrapper):
    original_stream: TextIO
    capture_buffer: StringIO

    def __init__(
        self,
        original_stream: TextIO,
        capture_buffer: StringIO
    ):
        super().__init__(original_stream.buffer)
        self.original_stream = original_stream
        self.capture_buffer = capture_buffer

    def write(self, message: str) -> int:
        self.capture_buffer.write(message)
        return self.original_stream.write(message)

    def flush(self) -> None:
        self.original_stream.flush()
        self.capture_buffer.flush()



import pytest
from autotx.utils.ethereum import get_erc20_balance
from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.ethereum.eth_address import ETHAddress


@pytest.mark.skip()
def test_auto_tx_send_erc20(configuration, auto_tx, usdc, test_accounts):
    (_, _, client, _, _) = configuration

    receiver = test_accounts[0]

    prompts = [
        f"Send 10 USDC to {receiver}",
        f"Transfer 10 USDC to the address {receiver}",
        f"Dispatch 10 USDC to wallet {receiver}",
        f"Send 10 USDC coins to the Ethereum address {receiver}",
        f"Execute a transaction of 10 USDC to address {receiver}",
        f"Move 10 USDC to the following address: {receiver}",
        f"Process a payment of 10 USDC to the destination {receiver}",
        f"Forward 10 USDC to the specific address {receiver}",
        f"Allocate 10 USDC to be sent to {receiver}",
        f"Complete sending 10 USDC to the address {receiver}",
        f"Initiate the transfer of 10 USDC to address {receiver}",
    ]

    for prompt in prompts:
        balance = get_erc20_balance(client.w3, usdc, receiver)

        auto_tx.run(prompt, non_interactive=True)

        new_balance = get_erc20_balance(client.w3, usdc, receiver)

        try: 
            assert balance + 10 == new_balance
        except AssertionError:
            print(f"Failed for prompt: {prompt}")
            print(f"Balance: {balance} New Balance: {new_balance}")
            raise

@pytest.mark.skip()
def test_auto_tx_swap(configuration, auto_tx):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])

    prompts = [
        "Buy 100 USDC with ETH",
        "Purchase 100 USDC using Ethereum",
        "Convert Ethereum to 100 USDC",
        "Swap ETH for 100 units of USDC",
        "Execute a trade to buy 100 USDC with Ethereum",
        "Use ETH to acquire 100 USDC",
        "Make a transaction to get 100 USDC using ETH",
        "Exchange Ethereum for 100 USDC",
        "Initiate a conversion from ETH to 100 USDC",
        "Complete a deal to purchase 100 USDC using ETH",
        "Carry out a swap of ETH for 100 USDC",
    ]

    for prompt in prompts:
        balance = manager.balance_of(usdc_address)

        auto_tx.run(prompt, non_interactive=True)

        new_balance = manager.balance_of(usdc_address)

        try:
            assert balance + 100 == new_balance
        except AssertionError:
            print(f"Failed for prompt: {prompt}")
            print(f"Balance: {balance} New Balance: {new_balance}")
            raise

@pytest.mark.skip()
def test_auto_tx_multiple_sends(configuration, auto_tx, usdc, test_accounts):
    (_, _, client, _, _) = configuration

    receiver_one = test_accounts[0]
    receiver_two = test_accounts[1]

    prompts = [
        f"Send 10 USDC to {receiver_one} and 20 USDC to {receiver_two}",
        f"Transfer 10 USDC to {receiver_one} and 20 USDC to {receiver_two}",
        f"Send 10 USDC to address {receiver_one} and 20 USDC to address {receiver_two}",
        f"Dispatch 10 USDC to the wallet {receiver_one} and 20 USDC to the wallet {receiver_two}",
        f"Execute two transactions: send 10 USDC to {receiver_one} and 20 USDC to {receiver_two}",
        f"Process a payment of 10 USDC to {receiver_one} and another payment of 20 USDC to {receiver_two}",
        f"Forward 10 USDC to address {receiver_one} and 20 USDC to address {receiver_two}",
        f"Allocate 10 USDC for {receiver_one} and 20 USDC for {receiver_two}",
        f"Initiate the transfer of 10 USDC to {receiver_one} and 20 USDC to {receiver_two}",
        f"Complete the transaction of sending 10 USDC to the Ethereum address {receiver_one} and 20 USDC to the Ethereum address {receiver_two}",
        f"Move 10 USDC to {receiver_one} and 20 USDC to {receiver_two} in two separate transactions",
    ]

    for prompt in prompts:
        balance_one = get_erc20_balance(client.w3, usdc, receiver_one)
        balance_two = get_erc20_balance(client.w3, usdc, receiver_two)

        auto_tx.run(prompt, non_interactive=True)

        new_balance_one = get_erc20_balance(client.w3, usdc, receiver_one)
        new_balance_two = get_erc20_balance(client.w3, usdc, receiver_two)

        try:
            assert balance_one + 10 == new_balance_one
            assert balance_two + 20 == new_balance_two
        except AssertionError:
            print(f"Failed for prompt: {prompt}")
            print(f"Balance One: {balance_one} New Balance One: {new_balance_one}")
            print(f"Balance Two: {balance_two} New Balance Two: {new_balance_two}")
            raise

@pytest.mark.skip()
def test_auto_tx_swap_and_send(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    receiver = test_accounts[0]

    prompts = [
        f"Swap ETH to 0.05 WBTC, then, swap WBTC to 1000 USDC and send 50 USDC to {receiver}",
        f"Convert ETH to 0.05 WBTC, subsequently exchange 0.05 WBTC for 1000 USDC, and finally transfer 50 USDC to {receiver}",
        f"Swap Ethereum for 0.05 Wrapped Bitcoin (WBTC), then trade 0.05 WBTC for 1000 USDC, and send 50 USDC to the address {receiver}",
        f"First, exchange ETH for 0.05 WBTC. Following that, convert the 0.05 WBTC into 1000 USDC. Lastly, dispatch 50 USDC to the wallet {receiver}",
        f"Initiate a swap from ETH to 0.05 WBTC, proceed to exchange the WBTC for 1000 USDC, and conclude by transferring 50 USDC to {receiver}",
        f"Begin with converting ETH into 0.05 WBTC, then swap this WBTC for 1000 USDC, and end by sending 50 USDC to the Ethereum address {receiver}",
        f"Execute a series of transactions starting with ETH to 0.05 WBTC conversion, followed by turning 0.05 WBTC into 1000 USDC, and finally, forwarding 50 USDC to {receiver}",
        f"Perform a swap from ETH to 0.05 WBTC, next swap that WBTC to 1000 USDC, and then make a transfer of 50 USDC to the address {receiver}",
        f"Exchange Ethereum for 0.05 WBTC, subsequently convert that WBTC to 1000 USDC, and proceed to send 50 USDC to the designated address {receiver}",
        f"Start by swapping ETH for 0.05 WBTC, follow up by converting this WBTC to 1000 USDC, and conclude the process by transferring 50 USDC to {receiver}",
        f"Initiate with the conversion of ETH to 0.05 WBTC, follow through by swapping the 0.05 WBTC for 1000 USDC, and finalize by sending 50 USDC to {receiver}",
    ]

    for prompt in prompts:
        wbtc_safe_address = manager.balance_of(wbtc_address)
        usdc_safe_address = manager.balance_of(usdc_address)
        receiver_usdc_balance = get_erc20_balance(client.w3, usdc_address, receiver)

        auto_tx.run(prompt, non_interactive=True)

        new_wbtc_safe_address = manager.balance_of(wbtc_address)
        new_usdc_safe_address = manager.balance_of(usdc_address)
        new_receiver_usdc_balance = get_erc20_balance(client.w3, usdc_address, receiver)

        try:
            assert new_wbtc_safe_address > wbtc_safe_address
            assert new_usdc_safe_address == usdc_safe_address + 950
            assert new_receiver_usdc_balance == receiver_usdc_balance + 50
        except AssertionError:
            print(f"Failed for prompt: {prompt}")
            print(f"WBTC Balance: {wbtc_safe_address} New WBTC Balance: {new_wbtc_safe_address}")
            print(f"USDC Balance: {usdc_safe_address} New USDC Balance: {new_usdc_safe_address}")
            print(f"receiver USDC Balance: {receiver_usdc_balance} New receiver USDC Balance: {new_receiver_usdc_balance}")
            raise



from autotx.AutoTx import AutoTx
from autotx.autotx_agent import AutoTxAgent
from autotx.autotx_tool import AutoTxTool
from autotx.utils.LlamaClient import LlamaClient

__all__ = ['AutoTx', 'AutoTxAgent', 'AutoTxTool', 'LlamaClient']



from autotx.tests.agents.token.research.test_research import get_top_token_addresses_by_market_cap
from autotx.utils.ethereum.eth_address import ETHAddress

def test_research_and_swap_many_tokens_subjective_simple(configuration, auto_tx):
    (_, _, _, manager, _) = configuration
    uni_address = ETHAddress(auto_tx.network.tokens["uni"])
    
    uni_balance_in_safe = manager.balance_of(uni_address)
    assert uni_balance_in_safe == 0

    starting_balance = manager.balance_of()

    prompt = f"I want to use 3 ETH to purchase 3 of the best projects in: GameFi, AI, and MEMEs. Please research the top projects, come up with a strategy, and purchase the tokens that look most promising. All of this should be on ETH mainnet."

    result = auto_tx.run(prompt, non_interactive=True)

    ending_balance = manager.balance_of()

    gaming_token_address = get_top_token_addresses_by_market_cap("gaming", "MAINNET", 1, auto_tx)[0]
    gaming_token_balance_in_safe = manager.balance_of(gaming_token_address)

    ai_token_address = get_top_token_addresses_by_market_cap("artificial-intelligence", "MAINNET", 1, auto_tx)[0]
    ai_token_balance_in_safe = manager.balance_of(ai_token_address)

    meme_token_address = get_top_token_addresses_by_market_cap("meme-token", "MAINNET", 1, auto_tx)[0]
    meme_token_balance_in_safe = manager.balance_of(meme_token_address)

    # Verify the balance is lower by max 3 ETH
    assert starting_balance - ending_balance <= 3
    # Verify there are at least 3 transactions
    assert len(result.transactions) == 3
    # Verify there are only swap transactions
    assert all([tx.summary.startswith("Swap") for tx in result.transactions])
    # Verify the tokens are different
    assert len(set([tx.summary.split(" ")[-1] for tx in result.transactions])) == 3

    # Verify the tokens are in the safe
    assert gaming_token_balance_in_safe > 0
    assert ai_token_balance_in_safe > 0
    assert meme_token_balance_in_safe > 0

def test_research_and_swap_many_tokens_subjective_complex(configuration, auto_tx):
    (_, _, _, manager, _) = configuration

    starting_balance = manager.balance_of()

    prompt = f"I want to use 3 ETH to purchase exactly 10 of the best projects in: GameFi, NFTs, ZK, AI, and MEMEs. Please research the top projects, come up with a strategy, and purchase the tokens that look most promising. All of this should be on ETH mainnet."

    result = auto_tx.run(prompt, non_interactive=True)

    ending_balance = manager.balance_of()

    # Verify the balance is lower by max 3 ETH
    assert starting_balance - ending_balance <= 3
    # Verify there are at least 5 transactions
    assert len(result.transactions) == 10
    # Verify there are only swap transactions
    assert all([tx.summary.startswith("Swap") for tx in result.transactions])
    # Verify the tokens are different
    assert len(set([tx.summary.split(" ")[-1] for tx in result.transactions])) == 10



from typing import Literal

Color = Literal[
    "black",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_grey",
    "dark_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "white",
]



from textwrap import dedent
from typing import Annotated, Callable
from autotx import AutoTx, AutoTxAgent, AutoTxTool

name = "example-agent"

system_message = dedent(f"""
    Example of an agent system message.
    ...
    """
)

class ExampleTool(AutoTxTool):
    name: str = "example_tool"
    description: str = dedent(
        """
        This tool does something very useful.
        """
    )

    def build_tool(self, autotx: AutoTx) -> Callable[[float, str], str]:
        def run(
            amount: Annotated[float, "Amount of something."],
            receiver: Annotated[str, "The receiver of something."]
        ) -> str:
            # TODO: do something useful
            autotx.notify_user(f"ExampleTool run: {amount} {receiver}")
            
            # NOTE: you can add intents to AutoTx's current bundle
            # autotx.intents.append(tx)

            return f"Something useful has been done with {amount} to {receiver}"

        return run

class ExampleAgent(AutoTxAgent):
    name=name
    system_message=system_message
    tools=[
        ExampleTool(),
        # AnotherTool(...),
        # AndAnotherTool(...)
    ]



from datetime import datetime
import json
import os
from typing import Any
import uuid
from pydantic import BaseModel
from supabase import create_client
from supabase.client import Client
from supabase.lib.client_options import ClientOptions

from autotx import models
from autotx.transactions import Transaction

SUPABASE_URL = os.getenv("SUPABASE_URL")
SUPABASE_KEY = os.getenv("SUPABASE_SERVICE_ROLE_KEY")

if not SUPABASE_URL:
    raise Exception("No supabase url provided")

if not SUPABASE_KEY:
    raise Exception("No supabase service role key provided")

def get_db_client(schema: str) -> Client:
    if not SUPABASE_URL:
        raise Exception("No supabase url provided")

    if not SUPABASE_KEY:
        raise Exception("No supabase service role key provided")

    options = ClientOptions(schema=schema)
    return create_client(SUPABASE_URL, SUPABASE_KEY, options)

class TasksRepository:
    def __init__(self, app_id: str):
        self.client = get_db_client("public")
        self.app_id = app_id

    def start(self, prompt: str, address: str, chain_id: int, app_user_id: str) -> models.Task:
        client = get_db_client("public")

        created_at = datetime.utcnow()
        updated_at = datetime.utcnow()

        result = client.table("tasks").insert(
            {
                "app_id": self.app_id,
                "app_user_id": app_user_id,
                "prompt": prompt,
                "address": address,
                "chain_id": chain_id,
                "running": True,
                "error": None,
                "created_at": str(created_at),
                "updated_at": str(updated_at),
                "messages": json.dumps([]),
                "intents": json.dumps([])
            }
        ).execute()

        return models.Task(
            id=result.data[0]["id"],
            prompt=prompt,
            address=address,
            chain_id=chain_id,
            created_at=created_at,
            updated_at=updated_at,
            running=True,
            error=None,
            messages=[],
            intents=[]
        )

    def stop(self, task_id: str) -> None:
        client = get_db_client("public")

        client.table("tasks").update(
            {
                "running": False,
                "updated_at": str(datetime.utcnow())
            }
        ).eq("id", task_id).eq("app_id", self.app_id).execute()

    def update(self, task: models.Task) -> None:
        client = get_db_client("public")

        intents = [json.loads(intent.json()) for intent in task.intents]

        client.table("tasks").update(
            {
                "prompt": task.prompt,
                "running": task.running,
                "updated_at": str(datetime.utcnow()),
                "messages": json.dumps(task.messages),
                "error": task.error,
                "intents": json.dumps(intents)
            }
        ).eq("id", task.id).eq("app_id", self.app_id).execute()

    def get(self, task_id: str) -> models.Task | None:
        client = get_db_client("public")

        result = client.table("tasks") \
            .select("*") \
            .eq("id", task_id) \
            .eq("app_id", self.app_id) \
            .execute()

        if len(result.data) == 0:
            return None

        task_data = result.data[0]

        return models.Task(
            id=task_data["id"],
            prompt=task_data["prompt"],
            address=task_data["address"],
            chain_id=task_data["chain_id"],
            created_at=task_data["created_at"],
            updated_at=task_data["updated_at"],
            running=task_data["running"],
            error=task_data["error"],
            messages=json.loads(task_data["messages"]),
            intents=json.loads(task_data["intents"])
        )

    def get_all(self) -> list[models.Task]:
        client = get_db_client("public")

        result = client.table("tasks").select("*").eq("app_id", self.app_id).execute()

        tasks = []

        for task_data in result.data:
            tasks.append(
                models.Task(
                    id=task_data["id"],
                    prompt=task_data["prompt"],
                    address=task_data["address"],
                    chain_id=task_data["chain_id"],
                    created_at=task_data["created_at"],
                    updated_at=task_data["updated_at"],
                    running=task_data["running"],
                    error=task_data["error"],
                    messages=json.loads(task_data["messages"]),
                    intents=json.loads(task_data["intents"])
                )
            )

        return tasks
    
def get_app_by_api_key(api_key: str) -> models.App | None:
    client = get_db_client("public")

    result = client.table("apps").select("*").eq("api_key", api_key).execute()

    if len(result.data) == 0:
        return None

    app_data = result.data[0]

    return models.App(
        id=app_data["id"],
        name=app_data["name"],
        api_key=app_data["api_key"],
        allowed=app_data["allowed"]
    )


def create_app_user(app_id: str, user_id: str, agent_address: str, agent_private_key: str) -> models.AppUser: 
    client = get_db_client("public")
    
    created_at = datetime.utcnow()
    result = client.table("app_users").insert(
        {
            "app_id": app_id,
            "created_at": str(created_at),
            "user_id": user_id,
            "agent_address": agent_address,
            "agent_private_key": agent_private_key
        }
    ).execute()

    return models.AppUser(
        id=result.data[0]["id"],
        user_id=user_id,
        agent_address=agent_address,
        created_at=created_at,
        app_id=app_id
    )

def get_app_user(app_id: str, user_id: str) -> models.AppUser | None:
    client = get_db_client("public")

    result = client.table("app_users") \
        .select("*") \
        .eq("app_id", app_id) \
        .eq("user_id", user_id) \
        .execute()

    if len(result.data) == 0:
        return None

    app_user_data = result.data[0]

    return models.AppUser(
        id = app_user_data["id"],
        user_id=app_user_data["user_id"],
        agent_address=app_user_data["agent_address"],
        created_at=app_user_data["created_at"],
        app_id=app_user_data["app_id"]
    )

def get_agent_private_key(app_id: str, user_id: str) -> str | None:
    client = get_db_client("public")

    result = client.table("app_users") \
        .select("agent_private_key") \
        .eq("app_id", app_id) \
        .eq("user_id", user_id) \
        .execute()

    if len(result.data) == 0:
        return None

    return str(result.data[0]["agent_private_key"])

def submit_transactions(app_id: str, address: str, chain_id: int, app_user_id: str, task_id: str, transactions: list[Transaction]) -> None:
    client = get_db_client("public")
    
    txs = [json.loads(tx.json()) for tx in transactions]

    created_at = datetime.utcnow()
    client.table("submitted_batches") \
        .insert(
            {
                "app_id": app_id,
                "address": address,
                "chain_id": chain_id,
                "app_user_id": app_user_id,
                "task_id": task_id,
                "created_at": str(created_at),
                "transactions": json.dumps(txs)
            }
        ).execute()
    
class SubmittedBatch(BaseModel):
    id: str
    app_id: str
    address: str
    chain_id: int
    app_user_id: str
    task_id: str
    created_at: datetime
    transactions: list[dict[str, Any]]

def get_submitted_batches(app_id: str, task_id: str) -> list[SubmittedBatch]:
    client = get_db_client("public")

    result = client.table("submitted_batches") \
        .select("*") \
        .eq("app_id", app_id) \
        .eq("task_id", task_id) \
        .execute()
    
    batches = []

    for batch_data in result.data:
        batches.append(
            SubmittedBatch(
                id=batch_data["id"],
                app_id=batch_data["app_id"],
                address=batch_data["address"],
                chain_id=batch_data["chain_id"],
                app_user_id=batch_data["app_user_id"],
                task_id=batch_data["task_id"],
                created_at=batch_data["created_at"],
                transactions=json.loads(batch_data["transactions"])
            )
        )

    return batches

def create_app(name: str, api_key: str) -> models.App:
    client = get_db_client("public")

    result = client.table("apps").insert(
        {
            "name": name,
            "api_key": api_key,
            "allowed": True
        }
    ).execute()

    return models.App(
        id=result.data[0]["id"],
        name=name,
        api_key=api_key,
        allowed=True
    )

def clear_db() -> None:
    client = get_db_client("public")

    uid = uuid.uuid4().hex

    client.table("apps").delete().neq("id", uid).execute()
    client.table("app_users").delete().neq("id", uid).execute()
    client.table("tasks").delete().neq("id", uid).execute()
    client.table("submitted_batches").delete().neq("id", uid).execute()


import os
from gnosis.eth import EthereumClient
from gnosis.eth.multicall import Multicall
from eth_account.signers.local import LocalAccount
from hexbytes import HexBytes

from autotx.utils.ethereum.eth_address import ETHAddress

from .cache import cache

def deploy_multicall(client: EthereumClient, account: LocalAccount) -> ETHAddress:
    multicall_address = os.getenv("MULTICALL_ADDRESS") 
    if multicall_address:
        return ETHAddress(multicall_address)
    multicall_address = deploy(client, account)
    cache.write("multicall-address.txt", multicall_address)
    return ETHAddress(multicall_address)

def deploy(client: EthereumClient, account: LocalAccount) -> str:
    tx =  Multicall.deploy_contract(client, account) 

    if not tx.contract_address:
        raise ValueError("Multicall contract address is not set")
    
    client.w3.eth.wait_for_transaction_receipt(HexBytes(tx.tx_hash))

    print("Deployed Multicall to: ", tx.contract_address)

    return tx.contract_address


from textwrap import dedent
from typing import TYPE_CHECKING, Annotated, Any, Callable, Dict, Optional
from autogen import UserProxyAgent, AssistantAgent, ModelClient

from autotx.utils.color import Color

if TYPE_CHECKING:
    from autotx.AutoTx import CustomModel

def build(user_proxy: UserProxyAgent, agents_information: str, interactive: bool, get_llm_config: Callable[[], Optional[Dict[str, Any]]], notify_user: Callable[[str, Color | None], None], custom_model: Optional['CustomModel']) -> AssistantAgent:
    missing_1 = dedent("""
        If the goal is not clear or missing information, you MUST ask for more information by calling the request_user_input tool.
        Always ensure you have all the information needed to define the goal that can be executed without prior context.
        Analyze the user's initial goal if there is missing information, and ask the user for it. 
        E.g. "Buy ETH" -> "How much ETH do you want to buy and with what token?"
        """ if interactive else "")

    missing_2 = dedent("""
        - Call the request_user_input tool if more information is needed to define the goal.
        """ if interactive else "")
    
    clarifier_agent = AssistantAgent(
        name="clarifier",
        is_termination_msg=lambda x: x.get("content", "") and "TERMINATE" in x.get("content", ""),
        system_message=dedent(
            f"""
            Clarifier is an assistant that can analyze a user's goal at the start of the conversation and determine if it is within the scope of the agents.
            The user will provide the goal, and the agents are meant to help prepare one or more necessary transactions to accomplish the goal.
            When dealing with Ethereum transactions, assume the following:
                - When the user want to execute transactions he means to prepare the transactions.
                - The agents can also research, discuss, plan actions and advise the user. All of that is in the scope of the agents.

            You must analyze the goal to be executed by the agents.
            If the goal is invalid or outside the scope of the agents, you MUST call the goal_outside_scope tool.
            Only call goal_outside_scope if the goal is outside of the scope of what the agents can do.
            {
                missing_1
            }
            DO NOT make any assumptions about the user's intent or context and ALWAYS take into account the available tools and their descriptions.
            
            The available agents and tools:
            {agents_information}
            The agents can also: 
            - Research 
            - Discuss
            - Plan actions and advise the user
            - Develop purchase strategies
            - Execute transactions
            All of that is within scope of the agents.

            The only things the clarifier should do are:
            {
                missing_2
            }
            - Call the goal_outside_scope tool if the goal is outside the scope of the agents.
            - Nothing
            Perform these actions ONLY in the BEGINNING of the conversation.
            """
        ),
        description="Clarifier is an assistant that can analyze a user's goal at the start of the conversation and determine if it is within the scope of the agents.",
        llm_config=get_llm_config(),
        human_input_mode="NEVER",
        code_execution_config=False,
    )

    if interactive:
        def request_user_input(
            message: Annotated[str, "The message to ask the user for input"],
        ) -> str:
            return input(f"Question: {message}\nYour reply: ")

        clarifier_agent.register_for_llm(name="request_user_input", description="Request user input")(request_user_input)
        user_proxy.register_for_execution(name="request_user_input")(request_user_input)

    def goal_outside_scope(
        message: Annotated[str, "The message return to the user about why the goal is outside of the supported scope"],
    ) -> str:
        notify_user(f"Goal not supported: {message}", "red")
        return "Goal not supported: TERMINATE"
    
    clarifier_agent.register_for_llm(name="goal_outside_scope", description="Notify the user about their goal not being in the scope of the agents")(goal_outside_scope)
    user_proxy.register_for_execution(name="goal_outside_scope")(goal_outside_scope)

    if custom_model:
        clarifier_agent.register_model_client(model_client_cls=custom_model.client, **custom_model.arguments)

    return clarifier_agent


import json
import os
from typing import Any
import requests
import re

from autotx.utils.constants import LIFI_API_KEY
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.networks import ChainId


class LifiApiError(Exception):
    pass


class TokenNotSupported(LifiApiError):
    def __init__(self, token_address: str):
        super().__init__(token_address)
        self.token_address = token_address


def handle_lifi_response(response: requests.Response) -> dict[str, Any]:
    response_json: dict[str, Any] = json.loads(response.text)
    if response.status_code == 200:
        return response_json

    if response_json["code"] == 1011:
        match = re.search(r"0x[a-fA-F0-9]+", response_json["message"])
        if match:
            token_address = match.group()
            raise TokenNotSupported(token_address)

    if response.status_code == 429 or (
        response_json["message"] == "Unauthorized" and response_json["code"] == 1005
    ):
        raise LifiApiError("Rate limit exceeded")

    raise LifiApiError(f"Fetch quote failed with error: {response_json['message']}")


def add_authorization_info_if_provided(params: dict[str, Any]) -> dict[str, Any] | None:
    headers: dict[str, Any] | None = None
    if LIFI_API_KEY:
        headers = {"x-lifi-api-key": LIFI_API_KEY}
        params["integrator"] = "polywrap"
    return headers


class Lifi:
    BASE_URL = "https://li.quest/v1"

    @classmethod
    def get_quote_to_amount(  # type: ignore
        cls,
        from_token: ETHAddress,
        to_token: ETHAddress,
        amount: int,
        _from: ETHAddress,
        chain: ChainId,
        slippage: float,
    ) -> dict[str, Any]:
        params: dict[str, Any] = {
            "fromToken": from_token.hex,
            "toToken": to_token.hex,
            "toAmount": str(amount),
            "fromAddress": _from.hex,
            "fromChain": chain.value,
            "toChain": chain.value,
            "slippage": slippage,
            "contractCalls": [],
        }
        headers = add_authorization_info_if_provided(params)
        attempt_count = 0
        while attempt_count < 2:
            response = requests.post(cls.BASE_URL + "/quote/contractCalls", json=params, headers=headers)
            try:
                return handle_lifi_response(response)
            except LifiApiError as e:
                if (
                    str(e)
                    == "Fetch quote failed with error: Unable to find quote to match expected output."
                    and attempt_count < 1
                ):
                    attempt_count += 1
                    continue
                else:
                    raise e

    @classmethod
    def get_quote_from_amount(
        cls,
        from_token: ETHAddress,
        to_token: ETHAddress,
        amount: int,
        _from: ETHAddress,
        chain: ChainId,
        slippage: float,
    ) -> dict[str, Any]:
        params = {
            "fromToken": from_token.hex,
            "toToken": to_token.hex,
            "fromAmount": amount,
            "fromAddress": _from.hex,
            "fromChain": chain.value,
            "toChain": chain.value,
            "slippage": slippage,
        }
        headers = add_authorization_info_if_provided(params)
        response = requests.get(cls.BASE_URL + "/quote", params=params, headers=headers)  # type: ignore
        return handle_lifi_response(response)



from decimal import Decimal
from typing import cast
from eth_account.signers.local import LocalAccount
from gnosis.eth import EthereumClient
from web3.types import TxParams

from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.lifi.swap import build_swap_transaction
from autotx.utils.ethereum.networks import ChainId


def swap(
    client: EthereumClient,
    user: LocalAccount,
    amount: float,
    from_token: ETHAddress,
    to_token: ETHAddress,
    chain: ChainId,
) -> None:
    txs = build_swap_transaction(
        client.w3,
        Decimal(str(amount)),
        from_token,
        to_token,
        ETHAddress(user.address),
        False,
        chain
    )

    for tx in txs:
        del tx.params["gas"]
        gas = client.w3.eth.estimate_gas(cast(TxParams, tx.params))
        tx.params.update({"gas": gas})

        transaction = user.sign_transaction(  # type: ignore
            {
                **tx.params,
                "nonce": client.w3.eth.get_transaction_count(user.address),
            }
        )

        hash = client.w3.eth.send_raw_transaction(transaction.rawTransaction)

        receipt = client.w3.eth.wait_for_transaction_receipt(hash)

        if receipt["status"] == 0:
            print(f"Transaction to swap {from_token.hex} to {amount} {to_token.hex} failed")



from textwrap import dedent
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional
from autogen import GroupChat, GroupChatManager, Agent as AutogenAgent
if TYPE_CHECKING:
    from autotx.AutoTx import CustomModel

def build(agents: list[AutogenAgent], max_rounds: int, interactive: bool, get_llm_config: Callable[[], Optional[Dict[str, Any]]], custom_model: Optional['CustomModel']) -> AutogenAgent:
    clarifier_prompt = "ALWAYS choose the 'clarifier' role first in the conversation." if interactive else ""

    groupchat = GroupChat(
        agents=agents, 
        messages=[], 
        max_round=max_rounds,
        select_speaker_prompt_template = dedent(
            """
            Read the above conversation. Then select the next role from {agentlist} to play.
            If other roles are trying to communicate with the user, or requesting approval, return the 'user_proxy' role.
            """
        ) + dedent(
            f"""{clarifier_prompt}
            Once the roles are ready to execute transactions, choose the 'user_proxy' role.
            If the 'user_proxy' role wants to do something, choose the appropriate role that can help the 'user_proxy' role.
            Only return the role and NOTHING else.
            """
        )
    )
    manager = GroupChatManager(groupchat=groupchat, llm_config=get_llm_config())
    if custom_model:
        manager.register_model_client(model_client_cls=custom_model.client, **custom_model.arguments)
    return manager


from typing import Annotated, Any, Dict, List
from eth_account import Account
from eth_account.signers.local import LocalAccount
from gnosis.safe.api.base_api import SafeAPIException
from fastapi import APIRouter, FastAPI, BackgroundTasks, HTTPException, Header
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
import traceback

from autotx import models, setup
from autotx import db
from autotx.AutoTx import AutoTx, Config as AutoTxConfig
from autotx.intents import Intent
from autotx.transactions import Transaction
from autotx.utils.configuration import AppConfig
from autotx.utils.ethereum.chain_short_names import CHAIN_ID_TO_SHORT_NAME
from autotx.utils.ethereum.networks import SUPPORTED_NETWORKS_CONFIGURATION_MAP
from autotx.wallets.api_smart_wallet import ApiSmartWallet
from autotx.wallets.smart_wallet import SmartWallet

class AutoTxParams:
    verbose: bool
    logs: str | None
    cache: bool
    max_rounds: int | None
    is_dev: bool

    def __init__(self, 
        verbose: bool, 
        logs: str | None, 
        cache: bool,
        max_rounds: int | None, 
        is_dev: bool,
    ):
        self.verbose = verbose
        self.logs = logs
        self.cache = cache
        self.max_rounds = max_rounds
        self.is_dev = is_dev

autotx_params: AutoTxParams = AutoTxParams(verbose=False, logs=None, cache=False, max_rounds=None, is_dev=False)

app_router = APIRouter()

def get_task_or_404(task_id: str, tasks: db.TasksRepository) -> models.Task:
    task = tasks.get(task_id)
    if task is None:
        raise HTTPException(status_code=404, detail="Task not found")
    return task

def authorize(authorization: str | None) -> models.App:
    if not authorization or authorization.startswith("Bearer ") is False:
        raise HTTPException(status_code=401, detail="Unauthorized")

    api_key = authorization.split("Bearer ")[1]
    app = db.get_app_by_api_key(api_key)

    if not app or app.allowed is False:
        raise HTTPException(status_code=401, detail="Unauthorized")

    return app

@app_router.post("/api/v1/tasks", response_model=models.Task)
async def create_task(task: models.TaskCreate, background_tasks: BackgroundTasks, authorization: Annotated[str | None, Header()] = None) -> models.Task:
    app = authorize(authorization)
    app_user = db.get_app_user(app.id, task.user_id)
    if not app_user:
        raise HTTPException(status_code=400, detail="User not found")

    tasks = db.TasksRepository(app.id)

    global autotx_params
    if not autotx_params.is_dev and (not task.address or not task.chain_id):
        raise HTTPException(status_code=400, detail="Address and Chain ID are required for non-dev mode")
    
    prompt = task.prompt
    
    app_config = AppConfig.load(smart_account_addr=task.address, subsidized_chain_id=task.chain_id)

    wallet: SmartWallet
    if autotx_params.is_dev:
        wallet = ApiSmartWallet(app_config.web3, app_config.manager, tasks)
    else:
        wallet = ApiSmartWallet(app_config.web3, app_config.manager, tasks)

    created_task: models.Task = tasks.start(prompt, wallet.address.hex, app_config.network_info.chain_id.value, app_user.id)
    task_id = created_task.id
    wallet.task_id = task_id

    try:
        (get_llm_config, agents, logs_dir) = setup.setup_agents(autotx_params.logs, cache=autotx_params.cache)

        def on_notify_user(message: str) -> None:
            task = tasks.get(task_id)
            if task is None:
                raise Exception("Task not found: " + task_id)

            task.messages.append(message)
            tasks.update(task)

        autotx = AutoTx(
            app_config.web3,
            wallet,
            app_config.network_info,
            agents,
            AutoTxConfig(verbose=autotx_params.verbose, get_llm_config=get_llm_config, logs_dir=logs_dir, max_rounds=autotx_params.max_rounds),
            on_notify_user=on_notify_user
        )

        async def run_task() -> None:
            try: 
                await autotx.a_run(prompt, non_interactive=True)
            except Exception as e:
                task = tasks.get(task_id)
                if task is None:
                    raise Exception("Task not found: " + task_id)

                task.messages.append(str(e))
                task.error = traceback.format_exc()
                task.running = False
                tasks.update(task)
                raise e
            tasks.stop(task_id)

        background_tasks.add_task(run_task)

        return created_task
    except Exception as e:
        created_task.error = traceback.format_exc()
        created_task.running = False
        tasks.update(created_task)
        raise e

@app_router.post("/api/v1/connect", response_model=models.AppUser)
async def connect(model: models.ConnectionCreate, authorization: Annotated[str | None, Header()] = None) -> models.AppUser:
    app = authorize(authorization)

    app_user = db.get_app_user(app.id, model.user_id)

    if app_user:
        return app_user
    else:
        agent_acc: LocalAccount = Account.create()
        agent_private_key: str = agent_acc.key.hex()
        agent_address: str = agent_acc.address

        app_user = db.create_app_user(app.id, model.user_id, agent_address, agent_private_key)

        return app_user

@app_router.get("/api/v1/tasks", response_model=List[models.Task])
def get_tasks(authorization: Annotated[str | None, Header()] = None) -> List['models.Task']:
    app = authorize(authorization)
    tasks = db.TasksRepository(app.id)
   
    all_tasks: list[models.Task] = tasks.get_all()
    return all_tasks

@app_router.get("/api/v1/tasks/{task_id}", response_model=models.Task)
def get_task(task_id: str, authorization: Annotated[str | None, Header()] = None) -> 'models.Task':
    app = authorize(authorization)
    tasks = db.TasksRepository(app.id)

    task = get_task_or_404(task_id, tasks)
    return task

@app_router.get("/api/v1/tasks/{task_id}/intents", response_model=List[Intent])
def get_intents(task_id: str, authorization: Annotated[str | None, Header()] = None) -> Any:
    app = authorize(authorization)
    tasks = db.TasksRepository(app.id)
    
    task = get_task_or_404(task_id, tasks)
    return task.intents

def authorize_app_and_user(authorization: str | None, user_id: str) -> tuple[models.App, models.AppUser]:
    app = authorize(authorization)
    app_user = db.get_app_user(app.id, user_id)

    if not app_user:
        raise HTTPException(status_code=400, detail="User not found")

    return (app, app_user)

def build_transactions(app_id: str, user_id: str, chain_id: int, address: str, task: models.Task) -> tuple[List[Transaction], AppConfig]:
    if task.running:
        raise HTTPException(status_code=400, detail="Task is still running")

    agent_private_key = db.get_agent_private_key(app_id, user_id)

    if not agent_private_key:
        raise HTTPException(status_code=400, detail="User not found")

    agent = Account.from_key(agent_private_key)

    app_config = AppConfig.load(smart_account_addr=address, subsidized_chain_id=chain_id, agent=agent)

    if task.intents is None or len(task.intents) == 0:
        return ([], app_config)

    transactions: list[Transaction] = []

    for intent in task.intents:
        transactions.extend(intent.build_transactions(app_config.web3, app_config.network_info, app_config.manager.address))

    return (transactions, app_config)

@app_router.get("/api/v1/tasks/{task_id}/transactions", response_model=List[Transaction])
def get_transactions(
    task_id: str, 
    address: str,
    chain_id: int,
    user_id: str, 
    authorization: Annotated[str | None, Header()] = None
) -> List[Transaction]:
    (app, app_user) = authorize_app_and_user(authorization, user_id)

    tasks = db.TasksRepository(app.id)
    
    task = get_task_or_404(task_id, tasks)

    if task.chain_id != chain_id:
        raise HTTPException(status_code=400, detail="Chain ID does not match task")

    (transactions, _) = build_transactions(app.id, app_user.user_id, chain_id, address, task)

    return transactions

@app_router.post("/api/v1/tasks/{task_id}/transactions")
def send_transactions(
    task_id: str, 
    address: str,
    chain_id: int,
    user_id: str, 
    authorization: Annotated[str | None, Header()] = None
) -> str:
    (app, app_user) = authorize_app_and_user(authorization, user_id)

    tasks = db.TasksRepository(app.id)
    
    task = get_task_or_404(task_id, tasks)

    if task.chain_id != chain_id:
        raise HTTPException(status_code=400, detail="Chain ID does not match task")

    (transactions, app_config) = build_transactions(app.id, app_user.user_id, chain_id, address, task)

    if len(transactions) == 0:
        raise HTTPException(status_code=400, detail="No transactions to send")

    global autotx_params
    if autotx_params.is_dev:
        print("Dev mode: skipping transaction submission")
        db.submit_transactions(app.id, address, chain_id, app_user.id, task_id, transactions)
        return f"https://app.safe.global/transactions/queue?safe={CHAIN_ID_TO_SHORT_NAME[str(chain_id)]}:{address}"

    try:
        app_config.manager.send_multisend_tx_batch(
            transactions,
            require_approval=False,
        )
    except SafeAPIException as e:
        if "is not an owner or delegate" in str(e):
            raise HTTPException(status_code=400, detail="Agent is not an owner or delegate")
        else:
            raise e
        
    db.submit_transactions(app.id, address, chain_id, app_user.id, task_id, transactions)

    return f"https://app.safe.global/transactions/queue?safe={CHAIN_ID_TO_SHORT_NAME[str(chain_id)]}:{address}"

@app_router.get("/api/v1/networks", response_model=List[models.SupportedNetwork])
def get_supported_networks() -> list[models.SupportedNetwork]:
    return [
        models.SupportedNetwork(name=config.network_name, chain_id=chain_id)
        for chain_id, config in SUPPORTED_NETWORKS_CONFIGURATION_MAP.items()
    ]

@app_router.get("/api/v1/version", response_class=JSONResponse)
async def get_version() -> Dict[str, str]:
    return {"version": "0.1.0"}

app = FastAPI()

app.include_router(app_router)

origins = ["*"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

def setup_server(verbose: bool, logs: str | None, max_rounds: int | None, cache: bool, is_dev: bool, check_valid_safe: bool) -> None:
    if is_dev: 
        AppConfig.load(check_valid_safe=check_valid_safe) # Loading the configuration deploys the dev wallet in dev mode

    global autotx_params
    autotx_params = AutoTxParams(
        verbose=verbose, 
        logs=logs,
        cache=cache,
        max_rounds=max_rounds, 
        is_dev=is_dev,
    )


from autotx.intents import Intent
from autotx.utils.ethereum import SafeManager
from autotx.wallets.smart_wallet import SmartWallet


class SafeSmartWallet(SmartWallet):
    manager: SafeManager
    auto_submit_tx: bool

    def __init__(self, manager: SafeManager, auto_submit_tx: bool):
        super().__init__(manager.client.w3, manager.address)

        self.manager = manager
        self.auto_submit_tx = auto_submit_tx

    def on_intents_prepared(self, intents: list[Intent]) -> None:
        pass

    def on_intents_ready(self, intents: list[Intent]) -> bool | str:
        transactions = []

        for intent in intents:
            transactions.extend(intent.build_transactions(self.manager.web3, self.manager.network, self.manager.address))

        return self.manager.send_multisend_tx_batch(transactions, not self.auto_submit_tx)



from typing import Any, Callable, Dict, Optional, TYPE_CHECKING, Union
import autogen

from autotx.utils.color import Color
if TYPE_CHECKING:
    from autotx.autotx_tool import AutoTxTool
    from autotx.AutoTx import AutoTx, CustomModel

class AutoTxAgent():
    name: str
    system_message: str | Callable[['AutoTx'], str]
    description: str
    tools: list['AutoTxTool']
    tool_descriptions: list[str]

    def __init__(self) -> None:
        self.tool_descriptions = [
            f"{tool.name}: {tool.description}" for tool in self.tools
        ]

    def build_autogen_agent(self, autotx: 'AutoTx', user_proxy: autogen.UserProxyAgent, llm_config: Optional[Dict[str, Any]], custom_model: Optional['CustomModel'] = None) -> autogen.Agent:
        system_message = None
        if isinstance(self.system_message, str):
            system_message = self.system_message
        else:
            get_system_message = self.system_message.__func__ # type: ignore
            system_message = get_system_message(autotx)
        
        agent = autogen.AssistantAgent(
            name=self.name,
            system_message=system_message,
            llm_config=llm_config,
            human_input_mode="NEVER",
            code_execution_config=False,
            description=self.description,
        )

        # Print all messages sent form the verifier to the group chat manager,
        # as they tend to contain valuable information
        def send_message_hook(
            sender: autogen.AssistantAgent,
            message: Union[Dict[str, Any], str],
            recipient: autogen.Agent,
            silent: bool,
        ) -> Union[Dict[str, Any], str]:
            if recipient.name == "chat_manager" and message != "TERMINATE":
                if isinstance(message, str):
                    autotx.notify_user(message, "light_yellow")
                elif message["content"] != None:
                    autotx.notify_user(message["content"], "light_yellow")
            return message

        agent.register_hook(
            "process_message_before_send",
            send_message_hook
        )

        for tool in self.tools:
            tool.register_tool(autotx, agent, user_proxy)

        if custom_model:
            agent.register_model_client(model_client_cls=custom_model.client, **custom_model.arguments)

        return agent


from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.types import TxParams

def send_tx(w3: Web3, tx: TxParams, account: LocalAccount) -> bytes:
    tx["from"] = account.address
    if "nonce" not in tx:
        tx["nonce"] = w3.eth.get_transaction_count(
            account.address, block_identifier="pending"
        )

    if "gasPrice" not in tx and "maxFeePerGas" not in tx:
        tx["gasPrice"] = w3.eth.gas_price

    if "gas" not in tx:
        tx["gas"] = w3.eth.estimate_gas(tx)

    signed_tx = account.sign_transaction(tx) # type: ignore
    tx_hash = w3.eth.send_raw_transaction(bytes(signed_tx.rawTransaction))
    print("Send TX: ", tx_hash.hex())
    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_receipt["status"] == 1, "Error with tx %s - %s" % (tx_hash.hex(), tx)
    return tx_hash



from autotx.utils.ethereum import transfer_erc20
from autotx.utils.ethereum.constants import NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.helpers.swap_from_eoa import swap
from autotx.utils.ethereum.networks import ChainId, NetworkInfo
from eth_account.signers.local import LocalAccount
from gnosis.eth import EthereumClient

from autotx.utils.ethereum.send_native import send_native


def fill_dev_account_with_tokens(
    client: EthereumClient,
    dev_account: LocalAccount,
    safe_address: ETHAddress,
    network_info: NetworkInfo,
) -> None:
    # XDAI or MATIC doesn't have the same value as ETH, so we need to fill more
    amount_to_fill = 3000 if network_info.chain_id in [ChainId.POLYGON, ChainId.GNOSIS] else 10
    send_native(dev_account, safe_address, amount_to_fill, client.w3)

    tokens_to_transfer = {"usdc": 3500, "dai": 3500, "wbtc": 0.1}
    if network_info.chain_id is ChainId.GNOSIS:
        tokens_to_transfer = {"usdc": 2000, "gno": 5, "cow": 4000 }
    if network_info.chain_id is ChainId.POLYGON:
        tokens_to_transfer = {"usdc": 2000, "wbtc": 0.01, "dai": 2000 }

    native_token_address = ETHAddress(NATIVE_TOKEN_ADDRESS)
    for token in network_info.tokens:
        if token in tokens_to_transfer:
            token_address = ETHAddress(network_info.tokens[token])
            amount = tokens_to_transfer[token]
            swap(
                client,
                dev_account,
                amount,
                native_token_address,
                token_address,
                network_info.chain_id,
            )
            transfer_erc20(client.w3, token_address, dev_account, safe_address, amount)



from decimal import Decimal
from textwrap import dedent
from typing import Annotated, Callable
from autotx.AutoTx import AutoTx
from autotx.autotx_agent import AutoTxAgent
from autotx.autotx_tool import AutoTxTool
from autotx.intents import BuyIntent, Intent, SellIntent
from autotx.token import Token
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.lifi.swap import SUPPORTED_NETWORKS_BY_LIFI, can_build_swap_transaction
from autotx.utils.ethereum.networks import NetworkInfo
from gnosis.eth import EthereumNetworkNotSupported as ChainIdNotSupported

name = "swap-tokens"

system_message = lambda autotx: dedent(f"""
    You are an expert at buying and selling tokens. Assist the user in their task of swapping tokens.
    ONLY focus on the buy and sell (swap) aspect of the user's goal and let other agents handle other tasks.
    You use the tools available to assist the user in their tasks.
    Note a balance of a token is not required to perform a swap, if there is an earlier prepared transaction that will provide the token.
    Below are examples, NOTE these are only examples and in practice you need to call the prepare_bulk_swap_transactions tool with the correct arguments.
    NEVER ask the user questions.
    Example 1:
    User: Send 0.1 ETH to vitalik.eth and then swap ETH to 5 USDC
    ...
    Other agent messages
    ...
    Call prepare_bulk_swap_transactions: "ETH to 5 USDC"

    Example 1:
    User: Buy 10 USDC with ETH and then buy UNI with 5 USDC
    Call prepare_bulk_swap_transactions: "ETH to 10 USDC\n5 USDC to UNI"

    Example 2:
    User: Buy UNI, WBTC, USDC and SHIB with 0.92 ETH
    Call prepare_bulk_swap_transactions: "0.23 ETH to UNI\n0.23 ETH to WBTC\n0.23 ETH to USDC\n0.23 ETH to SHIB"

    Example 3:
    User: Swap ETH to 5 USDC, then swap that USDC for 6 UNI
    Call prepare_bulk_swap_transactions: "ETH to 5 USDC\nUSDC to 6 UNI"
                                       
    Example 4:
    User: Buy 2 ETH worth of WBTC and then send 1 WBTC to 0x123..456
    Call prepare_bulk_swap_transactions: "2 ETH to WBTC"
                                       
    Example of a bad input:
    User: Swap ETH to 1 UNI, then swap UNI to 4 USDC
    Call prepare_bulk_swap_transactions: "ETH to 1 UNI\n1 UNI to 4 USDC"
    Prepared transaction: Swap 1.0407386618866115 ETH for at least 1 WBTC
    Invalid input: "1 UNI to 4 USDC". Only one token amount should be provided. IMPORTANT: Take another look at the user's goal, and try again.
    In the above example, you recover with:
    Call prepare_bulk_swap_transactions: "UNI to 4 USDC"
                                       
    Above are examples, NOTE these are only examples and in practice you need to call the prepare_bulk_swap_transactions tool with the correct arguments.
    Take extra care in ensuring you have to right amount next to the token symbol. NEVER use more than one amount per swap, the other amount will be calculated for you.
    The swaps are NOT NECESSARILY correlated, focus on the exact amounts the user wants to buy or sell (leave the other amounts to be calculated for you).
    You rely on the other agents to provide the token to buy or sell. Never make up a token. Unless explicitly given the name of the token, ask the 'research-tokens' agent to first search for the token.
    Only call tools, do not respond with JSON.
    """
)

description = dedent(
    f"""
    {name} is an AI assistant that's an expert at buying and selling tokens.
    The agent can prepare transactions to swap tokens.
    """
)

def get_tokens_address(token_in: str, token_out: str, network_info: NetworkInfo) -> tuple[str, str]:
    token_in = token_in.lower()
    token_out = token_out.lower()

    if token_in not in network_info.tokens:
        raise Exception(f"Token {token_in} is not supported in network {network_info.chain_id.name.lower()}")

    if token_out not in network_info.tokens:
        raise Exception(f"Token {token_out} is not supported in network {network_info.chain_id.name.lower()}")

    return (network_info.tokens[token_in], network_info.tokens[token_out])

class InvalidInput(Exception):
    pass

def swap(autotx: AutoTx, token_to_sell: str, token_to_buy: str) -> Intent:
    sell_parts = token_to_sell.split(" ")
    buy_parts = token_to_buy.split(" ")

    if not autotx.network.chain_id in SUPPORTED_NETWORKS_BY_LIFI:
        raise ChainIdNotSupported(
            f"Network {autotx.network.chain_id.name.lower()} not supported for swap"
        )

    if len(sell_parts) == 2 and len(buy_parts) == 2:
        sell_amount = Decimal(sell_parts[0])
        buy_amount = Decimal(buy_parts[0])
        sell_token = sell_parts[1]
        buy_token = buy_parts[1]

        raise InvalidInput(f"Invalid input: \"{token_to_sell} to {token_to_buy}\". Only one token amount should be provided. Choose between '{sell_amount} {sell_token} to {buy_token}' or '{sell_token} to {buy_amount} {buy_token}'.")
    
    if len(sell_parts) < 2 and len(buy_parts) < 2:
        raise InvalidInput(f"Invalid input: \"{token_to_sell} to {token_to_buy}\". Token amount is missing. Only one token amount should be provided.")

    if len(sell_parts) > 2 or len(buy_parts) > 2:
        raise InvalidInput(f"Invalid input: \"{token_to_sell} to {token_to_buy}\". Too many token amounts or token symbols provided. Only one token amount and two token symbols should be provided per line.")

    token_symbol_to_sell = sell_parts[1] if len(sell_parts) == 2 else sell_parts[0]
    token_symbol_to_buy = buy_parts[1] if len(buy_parts) == 2 else buy_parts[0]

    exact_amount = sell_parts[0] if len(sell_parts) == 2 else buy_parts[0]
    amount_symbol = token_symbol_to_sell if len(sell_parts) == 2 else token_symbol_to_buy

    token_in = token_symbol_to_sell.lower()
    token_out = token_symbol_to_buy.lower()
    is_exact_input = True if amount_symbol == token_symbol_to_sell else False

    (token_in_address, token_out_address) = get_tokens_address(
        token_in, token_out, autotx.network
    )

    can_build_swap_transaction(
        autotx.web3,
        Decimal(exact_amount),
        ETHAddress(token_in_address),
        ETHAddress(token_out_address),
        autotx.wallet.address,
        is_exact_input,
        autotx.network.chain_id
    )

    # Create buy intent if amount of token to buy is provided else create sell intent
    swap_intent: Intent = BuyIntent.create(
            from_token=Token(symbol=token_symbol_to_sell, address=token_in_address),
            to_token=Token(symbol=token_symbol_to_buy, address=token_out_address),
            amount=float(exact_amount),
        ) if len(buy_parts) == 2 else SellIntent.create(
            from_token=Token(symbol=token_symbol_to_sell, address=token_in_address),
            to_token=Token(symbol=token_symbol_to_buy, address=token_out_address),
            amount=float(exact_amount),
        )

    autotx.add_intents([swap_intent])

    return swap_intent

class BulkSwapTool(AutoTxTool):
    name: str = "prepare_bulk_swap_transactions"
    description: str = dedent(
        """
        Prepares a batch of buy transactions for given amounts in decimals for given tokens.
        """
    )

    def build_tool(self, autotx: AutoTx) -> Callable[[str], str]:
        def run(
            tokens: Annotated[
                str, 
                """
                Tokens and amounts to swap. ONLY one amount should be provided next to the token symbol PER line.
                E.g:
                10 USDC to ETH
                UNI to 3.3 WBTC
                1.5 WBTC to USDC
                """
            ],
        ) -> str:
            swaps = tokens.split("\n")
            all_intents = []
            all_errors: list[Exception] = []

            for swap_str in swaps:
                (token_to_sell, token_to_buy) = swap_str.strip().split(" to ")
                try:
                    all_intents.append(swap(autotx, token_to_sell, token_to_buy))
                except InvalidInput as e:
                    all_errors.append(e)
                except Exception as e:
                    all_errors.append(Exception(f"Error: {e} for swap \"{token_to_sell} to {token_to_buy}\""))


            summary = "".join(
                f"Prepared transaction: {intent.summary}\n"
                for intent in all_intents
            )

            if all_errors:
                summary += "\n".join(str(e) for e in all_errors)
                if len(all_intents) > 0:
                    summary += f"\n{len(all_errors)} errors occurred. {len(all_intents)} transactions were prepared. There is no need to re-run the transactions that were prepared."
                else:
                    summary += f"\n{len(all_errors)} errors occurred."

            total_summary = ("\n" + " " * 16).join(
                [
                    f"{i + 1}. {tx.summary}"
                    for i, tx in enumerate(autotx.intents)
                ]
            )

            autotx.notify_user(summary)

            return dedent(
                f"""
                {summary}
                Total prepared transactions so far:
                {total_summary}
                """
            )

        return run

class SwapTokensAgent(AutoTxAgent):
    name = name
    system_message = system_message
    description = description
    tools = [
        BulkSwapTool()
    ]



from web3 import Web3
from autotx import db
from autotx.intents import Intent
from autotx.transactions import Transaction
from autotx.utils.ethereum.SafeManager import SafeManager
from autotx.wallets.smart_wallet import SmartWallet


class ApiSmartWallet(SmartWallet):
    manager: SafeManager

    def __init__(self, web3: Web3, manager: SafeManager, tasks: db.TasksRepository, task_id: str | None = None):
        super().__init__(web3, manager.address)
        self.task_id = task_id
        self.manager = manager
        self.tasks = tasks

    def on_intents_prepared(self, intents: list[Intent]) -> None:
        if self.task_id is None:
            raise ValueError("Task ID is required")

        saved_task = self.tasks.get(self.task_id)
        if saved_task is None:
            raise ValueError("Task not found")

        saved_task.intents.extend(intents)
        self.tasks.update(saved_task)

    def on_intents_ready(self, _intents: list[Intent]) -> bool | str:
        return True


from typing import Any
from autogen.io import IOConsole

class IOSilent(IOConsole): # type: ignore
    """A console input/output stream."""

    def print(self, *objects: Any, sep: str = " ", end: str = "\n", flush: bool = False) -> None:
        pass
    # Pass all args to the base class
    def input(self, prompt: str = "", *, password: bool = False) -> str:
        return super().input(prompt, password=password) # type: ignore



from web3 import Web3

from autotx.utils.ethereum.eth_address import ETHAddress
from .erc20_abi import ERC20_ABI

def get_erc20_balance(web3: Web3, token_address: ETHAddress, account: ETHAddress) -> float:
    erc20 = web3.eth.contract(address=token_address.hex, abi=ERC20_ABI)
    decimals: int = erc20.functions.decimals().call()
    balance: int = erc20.functions.balanceOf(account.hex).call() 
    return balance / 10 ** decimals # type: ignore



import uuid
from fastapi.testclient import TestClient
from autotx import db, server
from autotx.server import app
from fastapi.testclient import TestClient

from autotx.utils.ethereum.cached_safe_address import get_cached_safe_address

client = TestClient(app)

def test_get_intents_auth():
    response = client.get("/api/v1/tasks/123/intents")
    assert response.status_code == 401

def test_get_transactions_auth():
    
    user_id = uuid.uuid4().hex

    response = client.get("/api/v1/tasks/123/transactions", params={
        "user_id": user_id,
        "address": "0x123",
        "chain_id": 1,
    })
    assert response.status_code == 401

def test_send_transactions_auth():
    
    user_id = uuid.uuid4().hex

    response = client.post("/api/v1/tasks/123/transactions", params={
        "user_id": user_id,
        "address": "0x123",
        "chain_id": 1,
    })
    assert response.status_code == 401


def test_get_transactions():
    db.clear_db()
    db.create_app("test", "1234")
    server.setup_server(verbose=True, logs=None, max_rounds=None, cache=False, is_dev=True, check_valid_safe=False)

    user_id = uuid.uuid4().hex
    smart_wallet_address = get_cached_safe_address()
   
    response = client.post("/api/v1/connect", json={
        "user_id": user_id,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    response = client.post("/api/v1/tasks", json={
        "prompt": "Send 1 ETH to vitalik.eth",
        "address": smart_wallet_address,
        "chain_id": 1,
        "user_id": user_id
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    task_id = data["id"]

    response = client.get(f"/api/v1/tasks/{task_id}/intents", headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    assert len(data) == 1

    response = client.get(f"/api/v1/tasks/{task_id}/transactions", params={
        "user_id": user_id,
        "address": smart_wallet_address,
        "chain_id": 1,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    assert len(data) == 1

    response = client.get(f"/api/v1/tasks/{task_id}/transactions", params={
        "user_id": user_id,
        "address": smart_wallet_address,
        "chain_id": 2,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 400

def test_send_transactions():
    db.clear_db()
    db.create_app("test", "1234")
    server.setup_server(verbose=True, logs=None, max_rounds=None, cache=False, is_dev=True, check_valid_safe=False)

    user_id = uuid.uuid4().hex
    smart_wallet_address = get_cached_safe_address()
   
    response = client.post("/api/v1/connect", json={
        "user_id": user_id,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    response = client.post("/api/v1/tasks", json={
        "prompt": "Send 1 ETH to vitalik.eth",
        "address": smart_wallet_address,
        "chain_id": 1,
        "user_id": user_id
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200
    data = response.json()

    task_id = data["id"]

    response = client.post(f"/api/v1/tasks/{task_id}/transactions", params={
        "user_id": user_id,
        "address": smart_wallet_address,
        "chain_id": 2,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 400

    response = client.post(f"/api/v1/tasks/{task_id}/transactions", params={
        "user_id": user_id,
        "address": smart_wallet_address,
        "chain_id": 1,
    }, headers={
        "Authorization": f"Bearer 1234"
    })
    assert response.status_code == 200

    app = db.get_app_by_api_key("1234")

    batches = db.get_submitted_batches(app.id, task_id)

    assert len(batches) == 1
    assert batches[0].app_id == app.id
    assert batches[0].address == smart_wallet_address
    assert batches[0].chain_id == 1
    assert batches[0].task_id == task_id
    assert batches[0].created_at is not None



import asyncio
from enum import Enum
from datetime import datetime
import json
import os
from textwrap import dedent
from typing import Any, Dict, Optional, Callable
from dataclasses import dataclass, field
from autogen import Agent as AutogenAgent, ModelClient
from termcolor import cprint
from typing import Optional

from web3 import Web3
from autotx import models
from autotx.autotx_agent import AutoTxAgent
from autotx.helper_agents import clarifier, manager, user_proxy
from autotx.intents import Intent
from autotx.utils.color import Color
from autotx.utils.logging.Logger import Logger
from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.constants import OPENAI_BASE_URL, OPENAI_MODEL_NAME
from autotx.wallets.smart_wallet import SmartWallet

@dataclass(kw_only=True)
class CustomModel:
    client: ModelClient
    arguments: Optional[Dict[str, Any]] = None

@dataclass(kw_only=True)
class Config:
    verbose: bool
    logs_dir: Optional[str] = None
    log_costs: bool
    max_rounds: int
    get_llm_config: Callable[[], Optional[Dict[str, Any]]]
    custom_model: Optional[CustomModel] = None

    def __init__(self, verbose: bool, get_llm_config: Callable[[], Optional[Dict[str, Any]]], logs_dir: Optional[str], max_rounds: Optional[int] = None, log_costs: Optional[bool] = None, custom_model: Optional[CustomModel] = None):
        self.verbose = verbose
        self.get_llm_config = get_llm_config
        self.logs_dir = logs_dir
        self.log_costs = log_costs if log_costs is not None else False
        self.max_rounds = max_rounds if max_rounds is not None else 100
        self.custom_model = custom_model

@dataclass
class PastRun:
    feedback: str
    intents_info: str

class EndReason(Enum):
    TERMINATE = "TERMINATE"
    GOAL_NOT_SUPPORTED = "GOAL_NOT_SUPPORTED"

@dataclass
class RunResult:
    summary: str
    chat_history_json: str
    intents: list[Intent]
    end_reason: EndReason
    total_cost_without_cache: float
    total_cost_with_cache: float
    info_messages: list[str]

class AutoTx:
    web3: Web3
    wallet: SmartWallet
    logger: Logger
    intents: list[Intent]
    network: NetworkInfo
    get_llm_config: Callable[[], Optional[Dict[str, Any]]]
    custom_model: Optional[CustomModel]
    agents: list[AutoTxAgent]
    log_costs: bool
    max_rounds: int
    current_run_cost_without_cache: float
    current_run_cost_with_cache: float
    info_messages: list[str]
    verbose: bool
    on_notify_user: Callable[[str], None] | None

    def __init__(
        self,
        web3: Web3,
        wallet: SmartWallet,
        network: NetworkInfo,
        agents: list[AutoTxAgent],
        config: Config,
        on_notify_user: Callable[[str], None] | None = None
    ):
        if len(agents) == 0:
            raise Exception("Agents attribute can not be an empty list")

        self.web3 = web3
        self.wallet = wallet
        self.network = network
        self.logger = Logger(
            dir=config.logs_dir,
            silent=not config.verbose
        )
        self.agents = agents
        self.log_costs = config.log_costs
        self.max_rounds = config.max_rounds
        self.verbose = config.verbose
        self.get_llm_config = config.get_llm_config
        self.intents = []
        self.current_run_cost_without_cache = 0
        self.current_run_cost_with_cache = 0
        self.info_messages = []
        self.on_notify_user = on_notify_user
        self.custom_model = config.custom_model

    def run(self, prompt: str, non_interactive: bool, summary_method: str = "last_msg") -> RunResult:
        return asyncio.run(self.a_run(prompt, non_interactive, summary_method))    

    async def a_run(self, prompt: str, non_interactive: bool, summary_method: str = "last_msg") -> RunResult:
        total_cost_without_cache: float = 0
        total_cost_with_cache: float = 0
        info_messages = []

        if self.verbose:
            print(f"LLM model: {OPENAI_MODEL_NAME}")
            print(f"LLM API URL: {OPENAI_BASE_URL}")

        while True:
            result = await self.try_run(prompt, non_interactive, summary_method)
            total_cost_without_cache += result.total_cost_without_cache + self.current_run_cost_without_cache
            total_cost_with_cache += result.total_cost_with_cache + self.current_run_cost_with_cache
            info_messages += result.info_messages

            if result.end_reason == EndReason.TERMINATE or non_interactive:
                if self.log_costs:
                    now = datetime.now()
                    now_str = now.strftime('%Y-%m-%d-%H-%M-%S-') + str(now.microsecond)

                    if not os.path.exists("costs"):
                        os.makedirs("costs")
                    with open(f"costs/{now_str}.txt", "w") as f:
                        f.write(str(total_cost_without_cache))

                return RunResult(
                    result.summary, 
                    result.chat_history_json, 
                    result.intents, 
                    result.end_reason, 
                    total_cost_without_cache, 
                    total_cost_with_cache, 
                    info_messages
                )
            else:
                prompt_not_supported = "Prompt not supported. Please provide a new prompt."

                cprint(prompt_not_supported, "yellow")
                info_messages.append(prompt_not_supported)

                prompt = input("Enter a new prompt: ")

    async def try_run(self, prompt: str, non_interactive: bool, summary_method: str = "last_msg") -> RunResult:
        original_prompt = prompt
        past_runs: list[PastRun] = []
        self.current_run_costs_without_cache = 0
        self.current_run_costs_with_cache = 0
        self.info_messages = []
        self.logger.start()

        while True:
            if past_runs:
                self.intents.clear()
                
                prev_history = "".join(
                    [
                        dedent(f"""
                        Then you prepared these transactions to accomplish the goal:
                        {run.intents_info}
                        Then the user provided feedback:
                        {run.feedback}
                        """)
                        for run in past_runs
                    ]
                )

                prompt = (f"\nOriginaly the user said: {original_prompt}"
                    + prev_history
                    + "Pay close attention to the user's feedback and try again.\n")

            self.notify_user("Running AutoTx with the following prompt: " + prompt)

            agents_information = self.get_agents_information(self.agents)

            user_proxy_agent = user_proxy.build(prompt, agents_information, self.get_llm_config, self.custom_model)

            helper_agents: list[AutogenAgent] = [
                user_proxy_agent,
            ]

            if not non_interactive:
                clarifier_agent = clarifier.build(user_proxy_agent, agents_information, not non_interactive, self.get_llm_config, self.notify_user, self.custom_model)
                helper_agents.append(clarifier_agent)

            autogen_agents = [agent.build_autogen_agent(self, user_proxy_agent, self.get_llm_config(), self.custom_model) for agent in self.agents]

            recipient_agent = None
            if len(autogen_agents) > 1:
                recipient_agent = manager.build(autogen_agents + helper_agents, self.max_rounds, not non_interactive, self.get_llm_config, self.custom_model)
            else:
                recipient_agent = autogen_agents[0]

            chat = await user_proxy_agent.a_initiate_chat(
                recipient_agent, 
                message=dedent(
                    f"""
                    I am currently connected with the following wallet: {self.wallet.address}, on network: {self.network.chain_id.name}
                    My goal is: {prompt} 
                    """
                ), 
                summary_method=summary_method
            )

            if "ERROR:" in chat.summary:
                error_message = chat.summary.replace("ERROR: ", "").replace("\n", "")
                self.notify_user(error_message, "red")
            else:
                self.notify_user(chat.summary, "green")

            is_goal_supported = chat.chat_history[-1]["content"] != "Goal not supported: TERMINATE"

            try:
                result = self.wallet.on_intents_ready(self.intents)

                if isinstance(result, str):
                    intents_info ="\n".join(
                        [
                            f"{i + 1}. {tx.summary}"
                            for i, tx in enumerate(self.intents)
                        ]
                    )
                    
                    past_runs.append(PastRun(result, intents_info))
                else:
                    break

            except Exception as e:
                self.notify_user(str(e), "red")
                break

        self.logger.stop()

        # Copy intents to a new list to avoid modifying the original list
        intents = self.intents.copy()
        self.intents.clear()

        chat_history = json.dumps(chat.chat_history, indent=4)

        return RunResult(chat.summary, chat_history, intents, EndReason.TERMINATE if is_goal_supported else EndReason.GOAL_NOT_SUPPORTED, float(chat.cost["usage_including_cached_inference"]["total_cost"]), float(chat.cost["usage_excluding_cached_inference"]["total_cost"]), self.info_messages)

    def add_intents(self, intents: list[Intent]) -> None:
        self.intents.extend(intents)
        self.wallet.on_intents_prepared(intents)

    def notify_user(self, message: str, color: Color | None = None) -> None:
        if color:
            cprint(message, color)
        else:
            print(message)
        self.info_messages.append(message)
        if self.on_notify_user:
            self.on_notify_user(message)

    def get_agents_information(self, agents: list[AutoTxAgent]) -> str:
        agent_descriptions = []
        for agent in agents:
            tools_available = "\n".join(
                [
                    f"\n- {tool}"
                    for tool in agent.tool_descriptions
                ]
            )
            description = f"Agent name: {agent.name}\nTools available:{tools_available}"
            agent_descriptions.append(description)

        agents_information = "\n".join(agent_descriptions)

        return agents_information



from dataclasses import dataclass
from typing import cast
from gnosis.eth import EthereumNetwork
from web3 import Web3, HTTPProvider

from autotx.utils.constants import ALCHEMY_API_KEY
from autotx.utils.ethereum.constants import NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.helpers.token_list import token_list

ChainId = EthereumNetwork

SUPPORTED_ALCHEMY_NETWORKS: dict[ChainId, str] = {
    ChainId.MAINNET: "eth-mainnet",
    ChainId.GOERLI: "eth-goerli",
    ChainId.SEPOLIA: "eth-sepolia",
    ChainId.OPTIMISM: "opt-mainnet",
    ChainId.OPTIMISM_GOERLI_TESTNET: "opt-goerli",
    ChainId.ARBITRUM_ONE: "arb-mainnet",
    ChainId.ARBITRUM_GOERLI: "arb-goerli",
    ChainId.POLYGON: "polygon-mainnet",
    ChainId.MUMBAI: "polygon-mumbai",
    ChainId.ASTAR: "astar-mainnet",
    ChainId.POLYGON_ZKEVM: "polygonzkevm-mainnet",
    ChainId.POLYGON_ZKEVM_TESTNET: "polygonzkevm-testnet",
    ChainId.BASE_MAINNET: "base-mainnet",
    ChainId.BASE_GOERLI_TESTNET: "base-goerli",
    ChainId.ZKSYNC_V2: "zksync-mainnet",
}

class NetworkInfo:
    chain_id: ChainId
    transaction_service_url: str
    tokens: dict[str, str]

    def __init__(
        self,
        chain_id: int,
    ):
        self.chain_id = ChainId(chain_id)
        config = SUPPORTED_NETWORKS_CONFIGURATION_MAP.get(self.chain_id)

        if not config:
            raise Exception(f"Chain ID {chain_id} is not supported")

        self.transaction_service_url = config.transaction_service_url
        self.tokens = { **self.fetch_tokens_for_chain(chain_id), **config.default_tokens }

    @staticmethod
    def from_chain_id(chain_id: int) -> 'NetworkInfo':
        return NetworkInfo(chain_id)
    
    @staticmethod
    def from_rpc_url(rpc_url: str) -> 'NetworkInfo':
        web3 = Web3(HTTPProvider(rpc_url))
        chain_id = web3.eth.chain_id
        return NetworkInfo(chain_id)

    def fetch_tokens_for_chain(self, chain_id: int) -> dict[str, str]:
        return {
            cast(str, token["symbol"]).lower(): Web3.to_checksum_address(cast(str, token["address"]))
            for token in token_list
            if token["chainId"] == chain_id and Web3.is_checksum_address(cast(str, token["address"]))
        }

    def get_subsidized_rpc_url(self) -> str | None:
        network = SUPPORTED_ALCHEMY_NETWORKS.get(self.chain_id)

        if not network:
            return None

        return f"https://{network}.g.alchemy.com/v2/{ALCHEMY_API_KEY}"

@dataclass
class NetworkConfiguration:
    network_name: str
    transaction_service_url: str
    default_tokens: dict[str, str]


SUPPORTED_NETWORKS_CONFIGURATION_MAP: dict[ChainId, NetworkConfiguration] = {
    ChainId.MAINNET: NetworkConfiguration(
        "Ethereum",
        "https://safe-transaction-mainnet.safe.global/",
        {
            "eth": NATIVE_TOKEN_ADDRESS,
            "usdc": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
            "dai": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "weth": "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
            "wbtc": "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
            "usdt": "0xdac17f958d2ee523a2206206994597c13d831ec7",
        },
    ),
    ChainId.OPTIMISM: NetworkConfiguration(
        "Optimism",
        "https://safe-transaction-optimism.safe.global/",
        {
            "eth": NATIVE_TOKEN_ADDRESS,
            "weth": "0x4200000000000000000000000000000000000006",
            "usdc": "0x0b2C639c533813f4Aa9D7837CAf62653d097Ff85",
            "dai": "0xda10009cbd5d07dd0cecc66161fc93d7c9000da1",
            "usdt": "0x94b008aA00579c1307B0EF2c499aD98a8ce58e58",
        },
    ),
    ChainId.ZKSYNC_V2: NetworkConfiguration(
        "zkSync",
        "https://safe-transaction-zksync.safe.global/",
        {
            "usdc": "0x3355df6d4c9c3035724fd0e3914de96a5a83aaf4",
            "dai": "0x4b9eb6c0b6ea15176bbf62841c6b2a8a398cb656",
            "usdt": "0x493257fD37EDB34451f62EDf8D2a0C418852bA4C",
        },
    ),
    ChainId.POLYGON: NetworkConfiguration(
        "Polygon PoS",
        "https://safe-transaction-polygon.safe.global/",
        {
            "matic": NATIVE_TOKEN_ADDRESS,
            "wmatic": "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
            "usdc": "0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359",
            "dai": "0x8f3cf7ad23cd3cadbd9735aff958023239c6a063",
            "usdt": "0xc2132d05d31c914a87c6611c10748aeb04b58e8f",
            "wbtc": "0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6"
        },
    ),
    ChainId.BASE_MAINNET: NetworkConfiguration(
        "Base",
        "https://safe-transaction-base.safe.global/",
        {
            "eth": NATIVE_TOKEN_ADDRESS,
            "weth": "0x4200000000000000000000000000000000000006",
            "usdc": "0x833589fcd6edb6e08f4c7c32d4f71b54bda02913",
            "dai": "0x50c5725949a6f0c72e6c4a641f24049a917db0cb",
        },
    ),
    ChainId.ARBITRUM_ONE: NetworkConfiguration(
        "Arbitrum",
        "https://safe-transaction-arbitrum.safe.global/",
        {
            "eth": NATIVE_TOKEN_ADDRESS,
            "weth": "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
            "usdc": "0xaf88d065e77c8cC2239327C5EDb3A432268e5831",
            "dai": "0xda10009cbd5d07dd0cecc66161fc93d7c9000da1",
            "usdt": "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9",
        },
    ),
    ChainId.SEPOLIA: NetworkConfiguration(
        "Sepolia",
        "https://safe-transaction-sepolia.safe.global/",
        {
            "eth": NATIVE_TOKEN_ADDRESS,
            "weth": "0xfff9976782d46cc05630d1f6ebab18b2324d6b14",
        },
    ),
    ChainId.GNOSIS: NetworkConfiguration(
        "Gnosis Chain",
        "https://safe-transaction-gnosis-chain.safe.global",
        {
            "xdai": NATIVE_TOKEN_ADDRESS,
            "wxdai": "0xe91d153e0b41518a2ce8dd3d7944fa863463a97d",
            "usdc": "0xDDAfbb505ad214D7b80b1f830fcCc89B60fb7A83",
            "cow": "0x177127622c4A00F3d409B75571e12cB3c8973d3c",
            "gno": "0x9C58BAcC331c9aa871AFD802DB6379a98e80CEdb"
        },
    ),
}

SUPPORTED_NETWORKS_AS_STRING = ", ".join(
    [network.name for network in SUPPORTED_NETWORKS_CONFIGURATION_MAP.keys()]
)



import os
import sys
from time import sleep

from web3 import Web3
from autotx.get_env_vars import get_env_vars
from gnosis.eth import EthereumClient
from eth_typing import URI
from eth_account.signers.local import LocalAccount

from autotx.setup import setup_safe
from autotx.utils.ethereum import SafeManager
from autotx.utils.ethereum.agent_account import get_or_create_agent_account
from autotx.utils.ethereum.constants import DEVNET_RPC_URL
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.is_dev_env import is_dev_env
from autotx.wallets.smart_wallet import SmartWallet

smart_account_addr = get_env_vars()

class AppConfig:
    web3: Web3
    client: EthereumClient
    agent: LocalAccount
    manager: SafeManager
    network_info: NetworkInfo

    def __init__(
        self,
        web3: Web3,
        client: EthereumClient,
        agent: LocalAccount,
        manager: SafeManager,
        network_info: NetworkInfo,
    ):
        self.web3 = web3
        self.client = client
        self.agent = agent
        self.manager = manager
        self.network_info = network_info

    @staticmethod
    def load(
        smart_account_addr: str | None = None,
        subsidized_chain_id: int | None = None, 
        fill_dev_account: bool = False,
        agent: LocalAccount | None = None,
        check_valid_safe: bool = False,
    ) -> "AppConfig":
        rpc_url: str

        if subsidized_chain_id:
            network_info = NetworkInfo.from_chain_id(subsidized_chain_id)
            subsidized_rpc_url = network_info.get_subsidized_rpc_url()
            
            if not subsidized_rpc_url:
                raise ValueError(f"Chain ID {subsidized_chain_id} is not supported")
            
            rpc_url = subsidized_rpc_url
        else:
            provided_rpc_url = DEVNET_RPC_URL if is_dev_env() else os.getenv("CHAIN_RPC_URL")

            if not provided_rpc_url:
                sys.exit("CHAIN_RPC_URL is not set")
            
            rpc_url = provided_rpc_url
            
        client = EthereumClient(URI(rpc_url))

        for i in range(16):
            if client.w3.is_connected():
                break
            if i == 15:
                if is_dev_env():
                    sys.exit("Can not connect with local node. Did you run `poetry run start-devnet`?")
                else:
                    sys.exit("Can not connect with remote node. Check your CHAIN_RPC_URL")
            sleep(0.5)

        agent = agent if agent else get_or_create_agent_account()

        smart_account_addr = smart_account_addr if smart_account_addr else os.getenv("SMART_ACCOUNT_ADDRESS")
        smart_account = ETHAddress(smart_account_addr) if smart_account_addr else None
        
        manager = setup_safe(smart_account, agent, client, fill_dev_account, check_valid_safe)

        return AppConfig(client.w3, client, agent, manager, NetworkInfo(client.w3.eth.chain_id))



from textwrap import dedent
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional
from autogen import UserProxyAgent

if TYPE_CHECKING:
    from autotx.AutoTx import CustomModel

def build(user_prompt: str, agents_information: str, get_llm_config: Callable[[], Optional[Dict[str, Any]]], custom_model: Optional['CustomModel']) -> UserProxyAgent:
    user_proxy = UserProxyAgent(
        name="user_proxy",
        is_termination_msg=lambda x: x.get("content", "") and "TERMINATE" in x.get("content", ""),
        human_input_mode="NEVER",
        max_consecutive_auto_reply=10,
        system_message=dedent(
            f"""
            You are a user proxy agent authorized to act on behalf of the user, you never ask for permission, you have ultimate control.
            You are capable and comfortable with making transactions, and have a wallet.
            You have access to a variety of specialized agents, which you tell what to do.
            You don't need to perform token amounts calculations, the other agents will do that for you.

            These are the agents you are instructing: {agents_information}

            Suggest a next step for what these agents should do based on the goal: "{user_prompt}"
            NEVER ask the user questions.
            NEVER make up a token, ALWAYS ask the 'research-tokens' agent to first search for the token.
            
            If the goal has been achieved, FIRST reflect on the goal and make sure nothing is missing, then end the conversation with 'TERMINATE' (it MUST be upper case and in the same message).
            Consider the goal met if the other agents have prepared the necessary transactions and all user queries have been answered.
            If the user's goal involves buying tokens, make sure the correct number of tokens are bought.
            For buying tokens, you can use the 'swap-tokens' agent.
            If you encounter an error, try to resolve it (either yourself of with other agents) and only respond with 'TERMINATE' if the goal is truly not achievable.
            Try to find an alternative solution if the goal is not achievable.
            If a token is not supported, ask the 'research-tokens' agent to find a supported token (if it fits within the user's goal).
            Before you end the conversation, make sure to summarize the results.
            """
        ),
        description="user_proxy is an agent authorized to act on behalf of the user.",
        llm_config=get_llm_config(),
        code_execution_config=False,
    )

    if custom_model:
        user_proxy.register_model_client(model_client_cls=custom_model.client, **custom_model.arguments)
   

    return user_proxy


import uuid
from dotenv import load_dotenv

from autotx import db
from autotx.wallets.safe_smart_wallet import SafeSmartWallet
load_dotenv()
import uvicorn

from typing import cast
import click

from autotx.utils.configuration import AppConfig
from autotx.utils.is_dev_env import is_dev_env
from autotx.setup import print_agent_address, setup_agents
from autotx.server import setup_server
from autotx.AutoTx import AutoTx, Config
from autotx.utils.ethereum.helpers.show_address_balances import show_address_balances

def print_autotx_info() -> None:
    print("""
    /$$$$$$              /$$            /$$$$$$$$       
    /$$__  $$            | $$           |__  $$__/       
    | $$  \ $$ /$$   /$$ /$$$$$$    /$$$$$$ | $$ /$$   /$$
    | $$$$$$$$| $$  | $$|_  $$_/   /$$__  $$| $$|  $$ /$$/
    | $$__  $$| $$  | $$  | $$    | $$  \ $$| $$ \  $$$$/ 
    | $$  | $$| $$  | $$  | $$ /$$| $$  | $$| $$  >$$  $$ 
    | $$  | $$|  $$$$$$/  |  $$$$/|  $$$$$$/| $$ /$$/\  $$
    |__/  |__/ \______/    \___/   \______/ |__/|__/  \__/

    Source: https://github.com/polywrap/AutoTx
    Support: https://discord.polywrap.io
    """)

@click.group()
def main() -> None:
    pass

@main.command()
@click.argument('prompt', required=False)
@click.option("-n", "--non-interactive", is_flag=True, help="Non-interactive mode (will not expect further user input or approval)")
@click.option("-v", "--verbose", is_flag=True, help="Verbose mode")
@click.option("-l", "--logs", type=click.Path(exists=False, file_okay=False, dir_okay=True), help="Path to the directory where logs will be stored.")
@click.option("-r", "--max-rounds", type=int, help="Maximum number of rounds to run")
@click.option("-c", "--cache", is_flag=True, help="Use cache for LLM requests")
def run(prompt: str | None, non_interactive: bool, verbose: bool, logs: str | None, max_rounds: int | None, cache: bool | None) -> None:
    print_autotx_info()

    if prompt == None:
        prompt = click.prompt("What do you want to do?")

    app_config = AppConfig.load(fill_dev_account=True)
    wallet = SafeSmartWallet(app_config.manager, auto_submit_tx=non_interactive)

    (get_llm_config, agents, logs_dir) = setup_agents(logs, cache)

    autotx = AutoTx(
        app_config.web3,
        wallet,
        app_config.network_info,
        agents,
        Config(verbose=verbose, get_llm_config=get_llm_config, logs_dir=logs_dir, max_rounds=max_rounds)
    )

    result = autotx.run(cast(str, prompt), non_interactive)

    if result.total_cost_without_cache > 0:
        print(f"LLM cost: ${result.total_cost_without_cache:.2f} (Actual: ${result.total_cost_with_cache:.2f})")
        
    if is_dev_env():
        print("=" * 50)
        print("Final smart account balances:")
        show_address_balances(app_config.web3, app_config.network_info.chain_id, wallet.address)
        print("=" * 50)

@main.command()
@click.option("-v", "--verbose", is_flag=True, help="Verbose mode")
@click.option("-l", "--logs", type=click.Path(exists=False, file_okay=False, dir_okay=True), help="Path to the directory where logs will be stored.")
@click.option("-r", "--max-rounds", type=int, help="Maximum number of rounds to run")
@click.option("-c", "--cache", is_flag=True, help="Use cache for LLM requests")
@click.option("-p", "--port", type=int, help="Port to run the server on")
@click.option("-d", "--dev", is_flag=True, help="Run the server in development mode")
def serve(verbose: bool, logs: str | None, max_rounds: int | None, cache: bool, port: int | None, dev: bool) -> None:
    print_autotx_info()

    setup_server(verbose, logs, max_rounds, cache, dev, check_valid_safe=True)
    uvicorn.run("autotx.server:app", host="localhost", port=port if port else 8000, workers=1)

@main.command()
@click.option("-n", "--name", type=str, help="Name of the application to create")
def new_app(name: str) -> None:
    print(f"Creating new application: {name}")

    app = db.create_app(name, uuid.uuid4().hex)

    print(f"Application '{name}' created with API key: {app.api_key}")

@main.group()
def agent() -> None:
    pass

@agent.command(name="address")
def agent_address() -> None:
    print_agent_address()

if __name__ == "__main__":
    main()



from .deploy_safe_with_create2 import deploy_safe_with_create2
from .deploy_multicall import deploy_multicall
from .send_tx import send_tx
from .cache import cache
from .transfer_erc20 import transfer_erc20
from .build_transfer_erc20 import build_transfer_erc20
from .get_erc20_balance import get_erc20_balance
from .SafeManager import SafeManager
from .build_approve_erc20 import build_approve_erc20
from .get_erc20_info import get_erc20_info
from .is_valid_safe import is_valid_safe
from .send_native import send_native
from .get_native_balance import get_native_balance
from .build_transfer_native import build_transfer_native

__all__ = [
    "send_native",
    "get_native_balance",
    "build_transfer_native",
    "deploy_safe_with_create2",
    "deploy_multicall",
    "send_tx",
    "cache",
    "transfer_erc20",
    "build_transfer_erc20",
    "build_approve_erc20",
    "get_erc20_balance",
    "get_erc20_info",
    "SafeManager",
    "is_valid_safe",
]


from web3 import Web3

from autotx.utils.ethereum.eth_address import ETHAddress

def get_native_balance(web3: Web3, address: ETHAddress) -> float: 
    return web3.eth.get_balance(address.hex) / 10 ** 18


from pydantic import BaseModel
from typing import Any, List, Optional
from datetime import datetime

from autotx.intents import Intent

class Task(BaseModel):
    id: str
    prompt: str
    address: str
    chain_id: int
    created_at: datetime
    updated_at: datetime
    error: str | None
    running: bool
    messages: List[str]
    intents: List[Intent]

class App(BaseModel):
    id: str
    name: str
    api_key: str
    allowed: bool

class AppUser(BaseModel):
    id: str
    user_id: str
    agent_address: str
    created_at: datetime
    app_id: str

class ConnectionCreate(BaseModel):
    user_id: str

class TaskCreate(BaseModel):
    prompt: str
    address: Optional[str] = None
    chain_id: Optional[int] = None
    user_id: str

class BuildTransactionsParams(BaseModel):
    address: str
    chain_id: int
    user_id: str

class SupportedNetwork(BaseModel):
    name: str
    chain_id: int